#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   config.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Outpost runtime configuration functions.
"""

# pylint: disable=R0914

import json
import os
import sys
from pathlib import Path
from typing import Any, Dict, List
from uuid import uuid4

import msgspec
import requests
from colorama import Fore
from InquirerPy import inquirer
from InquirerPy.base.control import Choice
from InquirerPy.validator import EmptyInputValidator

import datature.nexus.cli.consts as CONSTS
from datature.nexus import Client
from datature.nexus.cli import messages
from datature.nexus.cli.outpost import consts as OUTPOST_CONSTS
from datature.nexus.cli.outpost import messages as MESSAGES
from datature.nexus.cli.outpost.types import (
    OutpostDeviceConfig,
    OutpostDeviceConfiguration,
    OutpostDeviceRuntime,
)
from datature.nexus.config import REQUEST_TIME_OUT_SECONDS
from datature.nexus.error import Error

CONFIGURATION_FILE_INCLUSIONS = [
    "outpost-active-learning.yml",
    "outpost-logging.yml",
    "outpost-metrics.yml",
    "outpost-prediction.yml",
    "runtime-manager.yml",
]


def select_default_artifact(outpost_device_config: OutpostDeviceConfig) -> None:
    """Select default artifact for the configuration."""
    client = Client(
        secret_key=outpost_device_config.device_credentials.secret_key,
        endpoint=OUTPOST_CONSTS.NEXUS_ENDPOINT,
    )
    workspace_artifacts = client.list_artifacts()

    if not workspace_artifacts:
        print(messages.NO_ARTIFACTS_MESSAGE)
        sys.exit(1)

    artifact_choices = [
        Choice(name=f"{artifact.display_name} [{artifact.id[-6:]}]", value=artifact)
        for artifact in workspace_artifacts
    ]

    artifact = inquirer.fuzzy(
        message="Select a default artifact for this device:",
        choices=artifact_choices,
        max_height=CONSTS.INQUIRER_CHOICES_MAX_VISIBLE_ENTRIES,
        border=True,
    ).execute()

    outpost_device_config.default_artifact_id = artifact.id
    outpost_device_config.device_credentials.default_project_id = artifact.project_id


def get_sample_configuration(
    outpost_device_config: OutpostDeviceConfig,
) -> Dict[str, Any]:
    """Get sample configuration from the server."""
    response = requests.get(
        (
            f"{OUTPOST_CONSTS.NEXUS_ENDPOINT}/workspaces/ws_"
            f"{outpost_device_config.device_credentials.workspace_id}/metadata/outpost"
        ),
        headers={
            "Secret-Key": outpost_device_config.device_credentials.secret_key,
            "Content-Type": "application/json",
        },
        timeout=REQUEST_TIME_OUT_SECONDS,
    )
    return response.json()["sampleSpec"]


def create_configuration(
    outpost_device_config: OutpostDeviceConfig,
) -> OutpostDeviceConfiguration:
    """Create a new Outpost configuration."""
    configuration_id = str(uuid4()).replace("-", "")[-15:]

    configuration_name = (
        inquirer.text(
            message="Enter a descriptive name for this configuration:",
            mandatory=True,
            validate=EmptyInputValidator(),
        )
        .execute()
        .strip()
    )

    configuration_choices = [
        Choice(name="default", value="default"),
        Choice(name="Custom", value="custom"),
    ]
    configuration_spec_type = inquirer.fuzzy(
        message="Specify which configuration spec you want to use:",
        choices=configuration_choices,
        mandatory=True,
        max_height=CONSTS.INQUIRER_CHOICES_MAX_VISIBLE_ENTRIES,
        border=True,
    ).execute()

    if configuration_spec_type == "default":
        configuration_spec = get_sample_configuration(outpost_device_config)

    else:
        configuration_version = (
            inquirer.text(
                message="Enter the configuration version:",
                default="1.0.0",
                validate=EmptyInputValidator(),
            )
            .execute()
            .strip()
        )

        configuration_folder = (
            inquirer.filepath(
                message="Enter the path to the folder containing the configuration files:",
                validate=EmptyInputValidator(),
            )
            .execute()
            .strip()
        )

        configuration_folder = Path(os.path.expanduser(configuration_folder)).resolve()
        configuration_data = {}

        for file in configuration_folder.iterdir():
            if file.is_file() and file.name in CONFIGURATION_FILE_INCLUSIONS:
                with open(file, "r", encoding="utf-8") as config_file:
                    configuration_data[file.name] = config_file.read()

        if len(configuration_data) != len(CONFIGURATION_FILE_INCLUSIONS):
            print(
                MESSAGES.MISSING_CONFIGURATION_FILES_ERROR_MESSAGE.format(
                    ", ".join(CONFIGURATION_FILE_INCLUSIONS)
                )
            )
            sys.exit(1)

        configuration_spec = {
            "kind": "OutpostRuntime",
            "configVersion": configuration_version,
            "configData": configuration_data,
        }

    select_default_artifact(outpost_device_config)

    try:
        config_data = {
            "id": f"outpostconfig_{configuration_id}",
            "name": configuration_name,
            "spec": {
                "kind": configuration_spec["kind"],
                "configVersion": configuration_spec["configVersion"],
                "configData": configuration_spec["configData"],
                "configVariables": {
                    "project_id": {
                        "kind": "String",
                        "role": "ProjectId",
                        "default": outpost_device_config.device_credentials.default_project_id,
                    },
                    "artifact_id": {
                        "kind": "String",
                        "role": "ArtifactId",
                        "default": outpost_device_config.default_artifact_id,
                    },
                },
            },
        }

        response = requests.post(
            (
                f"{OUTPOST_CONSTS.NEXUS_ENDPOINT}/workspaces/ws_"
                f"{outpost_device_config.device_credentials.workspace_id}/outpost/configurations"
            ),
            headers={
                "Secret-Key": outpost_device_config.device_credentials.secret_key,
                "Content-Type": "application/json",
            },
            data=json.dumps(config_data),
            timeout=REQUEST_TIME_OUT_SECONDS,
        )
        response.raise_for_status()

    except requests.exceptions.ConnectionError as exc:
        raise Error(MESSAGES.SERVER_CONNECTION_ERROR_MESSAGE) from exc

    except requests.exceptions.HTTPError as exc:
        if response.status_code == 400:
            raise Error(MESSAGES.INSUFFICIENT_DEVICE_QUOTA_ERROR_MESSAGE) from exc

        if response.status_code == 402:
            raise Error(MESSAGES.FEATURE_NOT_AVAILABLE_ERROR_MESSAGE) from exc

        if response.status_code == 403:
            raise Error(
                MESSAGES.SERVER_FORBIDDEN_ERROR_MESSAGE.format(
                    outpost_device_config.device_credentials.workspace_id
                )
            ) from exc

        if response.status_code != 405:
            raise Error(
                MESSAGES.SERVER_UNKNOWN_ERROR_MESSAGE.format(response.text)
            ) from exc

    outpost_configuration = msgspec.json.decode(
        json.dumps(response.json()), type=OutpostDeviceConfiguration
    )

    print(f"{Fore.GREEN}✓{Fore.RESET} Configuration created successfully.")
    print(f"\n\tConfiguration ID: {outpost_configuration.id}")
    print(f"\tConfiguration Name: {outpost_configuration.name}")
    print(f"\tConfiguration Version: {outpost_configuration.spec.config_version}")
    print("\tConfiguration Files:")
    for file_name in outpost_configuration.spec.config_data.keys():
        print(f"\t\t{file_name}")
    print()

    return outpost_configuration


def list_configurations(
    outpost_device_config: OutpostDeviceConfig,
) -> List[OutpostDeviceConfiguration]:
    """List all Outpost configurations."""
    try:
        response = requests.get(
            (
                f"{OUTPOST_CONSTS.NEXUS_ENDPOINT}/workspaces/ws_"
                f"{outpost_device_config.device_credentials.workspace_id}/outpost/configurations"
            ),
            headers={
                "Secret-Key": outpost_device_config.device_credentials.secret_key,
                "Content-Type": "application/json",
            },
            timeout=REQUEST_TIME_OUT_SECONDS,
        )
        response.raise_for_status()

    except requests.exceptions.ConnectionError as exc:
        raise Error(MESSAGES.SERVER_CONNECTION_ERROR_MESSAGE) from exc

    except requests.exceptions.HTTPError as exc:
        if response.status_code == 400:
            raise Error(MESSAGES.INSUFFICIENT_DEVICE_QUOTA_ERROR_MESSAGE) from exc

        if response.status_code == 402:
            raise Error(MESSAGES.FEATURE_NOT_AVAILABLE_ERROR_MESSAGE) from exc

        if response.status_code == 403:
            raise Error(
                MESSAGES.SERVER_FORBIDDEN_ERROR_MESSAGE.format(
                    outpost_device_config.device_credentials.workspace_id
                )
            ) from exc

        if response.status_code != 405:
            raise Error(
                MESSAGES.SERVER_UNKNOWN_ERROR_MESSAGE.format(response.text)
            ) from exc

    outpost_configurations = msgspec.json.decode(
        json.dumps(response.json()["data"]), type=List[OutpostDeviceConfiguration]
    )

    return outpost_configurations


def prompt_for_configuration_version(
    outpost_device_config: OutpostDeviceConfig,
) -> None:
    """Prompt for configuration version."""
    configurations = list_configurations(outpost_device_config)

    has_existing_configurations = len(configurations) > 0

    configuration_choices = [
        (
            Choice(name="Select an Existing Configuration", value="existing")
            if has_existing_configurations
            else None
        ),
        Choice(name="Register a New Configuration", value="new"),
    ]
    configuration_choices = [choice for choice in configuration_choices if choice]

    configuration_type = inquirer.select(
        message="Select deployment configuration type:",
        choices=configuration_choices,
        default="new" if not has_existing_configurations else "existing",
        border=True,
    ).execute()

    if configuration_type == "new":
        configuration = create_configuration(outpost_device_config)

    else:
        configuration_choices = [
            Choice(
                name=f"{configuration.name} (v{configuration.spec.config_version})",
                value=configuration,
            )
            for configuration in sorted(
                configurations, key=lambda c: c.create_date, reverse=True
            )
        ]

        configuration = inquirer.fuzzy(
            message="Select a configuration to deploy:",
            choices=configuration_choices,
            max_height=CONSTS.INQUIRER_CHOICES_MAX_VISIBLE_ENTRIES,
            border=True,
        ).execute()

    outpost_device_config.configuration = configuration
    outpost_device_config.write_to_file()


def prompt_for_runtime_config(outpost_device_config: OutpostDeviceConfig) -> None:
    """Prompt for runtime configuration."""
    runtimes_response = requests.get(
        (
            f"{OUTPOST_CONSTS.NEXUS_ENDPOINT}/workspaces/ws_"
            f"{outpost_device_config.device_credentials.workspace_id}"
            f"/outpost/runtimes"
        ),
        headers={
            "Secret-Key": outpost_device_config.device_credentials.secret_key,
            "Content-Type": "application/json",
        },
        timeout=REQUEST_TIME_OUT_SECONDS,
    )

    available_runtimes = msgspec.json.decode(
        json.dumps(runtimes_response.json()["data"]), type=List[OutpostDeviceRuntime]
    )

    sorted_runtimes = sorted(
        available_runtimes, key=lambda r: r.create_date, reverse=True
    )

    runtime_choices = []
    if sorted_runtimes:
        runtime_choices.append(
            Choice(
                name=f"latest (v{str(sorted_runtimes[0].version)}, recommended)",
                value=sorted_runtimes[0],
            )
        )
        runtime_choices.extend(
            [
                Choice(
                    name=f"v{str(runtime.version)}",
                    value=runtime,
                )
                for runtime in sorted_runtimes[1:]
            ]
        )

    runtime = inquirer.fuzzy(
        message="Select a runtime version to deploy:",
        choices=runtime_choices,
        default="latest",
        max_height=CONSTS.INQUIRER_CHOICES_MAX_VISIBLE_ENTRIES,
        border=True,
    ).execute()

    runtime_dir = inquirer.filepath(
        message="Specify destination folder for runtime files:",
        long_instruction=(
            "The system will create a `/runtime` subfolder at your specified location. "
            "If you leave this field empty, files will be installed in the default location:"
            f" `{OUTPOST_CONSTS.OUTPOST_ROOT_DIR}`."
        ),
        mandatory=True,
        only_directories=True,
        default=str(OUTPOST_CONSTS.OUTPOST_ROOT_DIR),
        validate=lambda result: len(result) > 0,
    ).execute()

    # convert to absolute path
    runtime_dir = Path(os.path.expanduser(runtime_dir)).resolve()

    outpost_device_config.runtime = runtime
    outpost_device_config.runtime_dir = str(runtime_dir)
    outpost_device_config.write_to_file()



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   config.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Configuration for local deployment
"""

# pylint: disable=R0902

from __future__ import annotations

import json
import os
from typing import List, Union

import msgspec
from msgspec import MsgspecError, Struct, field

from datature.nexus.cli.deploy.local.consts import (
    DEFAULT_ALLOWED_ORIGINS,
    DEFAULT_DEVICES,
    DEFAULT_PORT,
    DEFAULT_RUNTIME,
    DEPLOY_CONFIGS_FILE_PATH,
)


class DeploymentConfig(Struct, rename="camel", kw_only=True):
    """Configuration for local deployment"""

    deployment_id: str
    secret_key: str
    project_id: str
    project_name: str
    artifact_id: str
    artifact_name: str
    runtime: str = DEFAULT_RUNTIME
    devices: Union[str, int] = DEFAULT_DEVICES
    port: int = DEFAULT_PORT
    allowed_origins: List[str] = field(default_factory=lambda: DEFAULT_ALLOWED_ORIGINS)


class DeploymentConfigs(Struct, rename="camel", kw_only=True):
    """Configuration list for local deployment"""

    configs: List[DeploymentConfig] = field(default_factory=list)

    def __post_init__(self):
        if not DEPLOY_CONFIGS_FILE_PATH.exists():
            os.makedirs(DEPLOY_CONFIGS_FILE_PATH.parent, exist_ok=True)
            with open(DEPLOY_CONFIGS_FILE_PATH, "w", encoding="utf-8") as config_file:
                config_file.write("[]")

    def write_to_file(self) -> None:
        """Write DeploymentConfigs to file."""
        try:
            with open(DEPLOY_CONFIGS_FILE_PATH, "w", encoding="utf-8") as config_file:
                json.dump(msgspec.to_builtins(self), config_file)

        except (IOError, MsgspecError) as exc:
            print(
                f"Error writing DeploymentConfigs to {DEPLOY_CONFIGS_FILE_PATH}: {exc}"
            )
            raise exc

    @classmethod
    def read_from_file(cls) -> DeploymentConfigs:
        """Read DeploymentConfigs from file."""
        if not DEPLOY_CONFIGS_FILE_PATH.exists():
            return cls()

        with open(DEPLOY_CONFIGS_FILE_PATH, "r", encoding="utf-8") as config_file:
            json_data = json.loads(config_file.read())
            return msgspec.convert(json_data, type=DeploymentConfigs)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   inference.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Inference utility types.
"""

from dataclasses import dataclass
from enum import Enum
from typing import Tuple


class ModelExtension(Enum):
    """Model extension enum."""

    TF = "saved_model"
    TFLITE = ".tflite"
    ONNX = ".onnx"
    PYTORCH = ".pth"

    @classmethod
    def has_value(cls, value):
        """Check if enum contains value.

        Args:
            value: Value to check.

        Returns:
            True if value is in enum, False otherwise.
        """
        return value in cls._value2member_map_


class TensorDtype(Enum):
    """Data type of a tensor."""

    FP16 = "FP16"
    FP32 = "FP32"
    FP64 = "FP64"
    UINT8 = "UINT8"
    UINT16 = "UINT16"
    UINT32 = "UINT32"
    UINT64 = "UINT64"
    INT8 = "INT8"
    INT16 = "INT16"
    INT32 = "INT32"
    INT64 = "INT64"
    BOOL = "BOOL"


class PredictionTaskType(Enum):
    """Inference task type enum."""

    IMAGE_CLASSIFICATION = "ImageClassification"
    OBJECT_DETECTION = "ObjectDetection"
    INSTANCE_SEGMENTATION = "InstanceSegmentation"
    SEMANTIC_SEGMENTATION = "SemanticSegmentation"
    SEMANTIC_3D_SEGMENTATION = "Semantic3DSegmentation"
    KEYPOINT_DETECTION = "KeypointDetection"


@dataclass
class TensorDescription:
    """Description of a tensor returned by or fed into a model."""

    name: str
    dtype: TensorDtype
    shape: Tuple[int]



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   prediction_response.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Prediction response type.
"""

from typing import Any, Dict, List, Optional, Union

from msgspec import Struct


class Tag(Struct, kw_only=True, omit_defaults=True, rename="camel"):
    """
    Tag object.

    Attributes:
        name: The name of the tag.
        id: The id of the tag.
        color: The color of the tag.
    """

    name: str
    id: Optional[int] = None
    color: Optional[str] = None


class Medical3DImage(Struct, kw_only=True, rename="camel"):
    """
    Image object.

    Attributes:
        encoding: The base64 encoding of the image.
        file: The file name of the image.
    """

    encoding: str = "NiiGz"
    file: str


class BitmaskCOCORLE(Struct, kw_only=True, rename="camel"):
    """
    COCORLE mask object.

    Attributes:
        encoding: The encoding of the mask.
        height: The height of the mask.
        width: The width of the mask.
        counts: The counts of the mask.
    """

    encoding: str = "CocoRleMask"
    height: int
    width: int
    counts: Union[List[int], str]


class BitmaskNiiGz(Struct, kw_only=True, rename="camel"):
    """
    Bitmask 3D object.
    """

    encoding: str = "NiiGz"
    file: str


Bitmask2D = BitmaskCOCORLE
Bitmask3D = BitmaskNiiGz


class Prediction(Struct, kw_only=True, omit_defaults=True, rename="camel"):
    """
    Prediction object.

    Attributes:
        annotation_id: The id of the annotation.
        confidence: The confidence of the prediction.
        tag: The tag of the prediction.
        bound: The bound of the prediction.
        bound_type: The type of the bound.
        contour: The contour of the prediction.
        contour_type: The type of the contour.
        skeleton: The skeleton of the prediction.
        keypoints: The keypoints of the prediction.
        keypoint_confidences: The keypoint confidences of the prediction.
        bitmask: The bitmask of the prediction.
    """

    annotation_id: int
    confidence: float
    tag: Tag
    bound: Optional[List[List[float]]] = None
    bound_type: Optional[str] = None
    contour: Optional[List[List[float]]] = None
    contour_type: Optional[str] = None
    skeleton: Optional[Dict[str, Any]] = None
    keypoints: Optional[List[List[float]]] = None
    keypoint_confidences: Optional[List[float]] = None
    bitmask: Optional[Union[Bitmask2D, Bitmask3D]] = None


class PredictionResponse(Struct, kw_only=True, omit_defaults=True, rename="camel"):
    """
    Prediction response object.

    Attributes:
        predictions: The predictions of the response.
        avg_entropy: The average entropy of the response.
        avg_entropy_for_class: The average entropy for each class of the response.
        image: The image of the response.
        warnings: The warnings of the response.
    """

    predictions: Optional[List[Prediction]] = None
    avg_entropy: Optional[float] = None
    avg_entropy_for_class: Optional[Dict[str, float]] = None
    image: Optional[Medical3DImage] = None
    warnings: Optional[List[str]] = None


class APIDeploymentPredictionResponse(
    Struct, kw_only=True, omit_defaults=True, rename="camel"
):
    """
    API deployment prediction response object.

    Attributes:
        tags: The tags of the response.
        response: The response of the prediction.
    """

    tags: Dict[str, Tag]
    response: PredictionResponse



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   endpoints.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Custom endpoints for the inference server.
"""

import re
import time
from datetime import datetime
from typing import Tuple

import msgspec
from fastapi.routing import APIRoute
from litserve.server import LitServer
from starlette.responses import RedirectResponse

from datature.nexus.cli.deploy.local.logger import LOG_STREAM, LogStreamTracker
from datature.nexus.cli.deploy.local.types.endpoint_response import (
    HealthResponse,
    LogEntry,
    LogResponse,
)

log_filter_patterns = [
    "GET /health",
    "GET /logs",
]


def remove_terminal_colors(text: str) -> str:
    """Remove ANSI terminal color codes from a string."""
    ansi_escape_pattern = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
    return ansi_escape_pattern.sub("", text)


def parse_log_line(log_line: str) -> Tuple[int, str]:
    """
    Parse a log line into timestamp and message.

    Args:
        log_line: Log line in format "timestamp message"

    Returns:
        Tuple of (timestamp: int, message: str)
    """
    # split on the second space
    parts = log_line.split("\t", 1)

    if len(parts) == 2:
        try:
            timestamp = int(
                datetime.strptime(parts[0], "%Y-%m-%d %H:%M:%S.%f").timestamp() * 1000
            )
            message = parts[1]
            return timestamp, message
        except ValueError:
            return 0, log_line
    else:
        return 0, log_line


def add_custom_endpoints(server: LitServer, project_id: str, project_name: str):
    """Add custom endpoints to LitServer"""

    @server.app.get("/health")
    async def health(project_id: str = project_id, project_name: str = project_name):
        """Health check endpoint."""
        response = HealthResponse(
            status="ok",
            timestamp=int(time.time() * 1000),
            project=project_id,
            project_name=project_name,
        )
        return msgspec.to_builtins(response)

    @server.app.get("/logs")
    async def get_logs():
        """Get recent server logs."""
        log_lines = [
            remove_terminal_colors(line.strip())
            for line in reversed(
                LOG_STREAM.getvalue()[LogStreamTracker.pointer :].split("\n")
            )
            if line.strip()
            and not any(pattern in line.strip() for pattern in log_filter_patterns)
        ]

        LogStreamTracker.pointer = len(LOG_STREAM.getvalue())

        response = LogResponse(
            logs=[
                LogEntry(timestamp=timestamp, message=message)
                for timestamp, message in map(parse_log_line, log_lines)
            ]
        )

        return msgspec.to_builtins(response)

    @server.app.get("/all-logs")
    async def get_all_logs():
        """Get all historical server logs."""
        log_lines = [
            remove_terminal_colors(line.strip())
            for line in reversed(LOG_STREAM.getvalue().split("\n"))
            if line.strip()
        ]

        response = LogResponse(
            logs=[
                LogEntry(timestamp=timestamp, message=message)
                for timestamp, message in map(parse_log_line, log_lines)
            ]
        )

        return msgspec.to_builtins(response)

    def replace_route(path: str, new_handler, methods=None):
        if methods is None:
            methods = ["GET"]

        for i, route in enumerate(server.app.router.routes):
            if isinstance(route, APIRoute) and route.path == path:
                new_route = APIRoute(path=path, endpoint=new_handler, methods=methods)
                server.app.router.routes[i] = new_route
                break

    async def index():
        """Index endpoint."""
        return RedirectResponse(url="/docs")

    replace_route("/", index)

    return server



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   management_api.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Management API client.
"""

import os

from datature.nexus import Client
from datature.nexus.api.project import Project


class ManagementAPIClient:
    """Management API client."""

    _endpoint: str
    _secret_key: str
    _project_id: str
    _client: Client

    def __init__(self, secret_key: str, project_id: str) -> None:
        self._endpoint = os.getenv("DATATURE_API_BASE_URL", "https://api.datature.io")
        self._secret_key = secret_key
        self._project_id = project_id

        self._client = Client(secret_key=self._secret_key, endpoint=self._endpoint)

    @property
    def project(self) -> Project:
        """Get the project."""
        return self.client.get_project(self._project_id)

    @property
    def project_id(self) -> str:
        """Get the project ID."""
        return self._project_id

    @project_id.setter
    def project_id(self, project_id: str) -> None:
        """Set the project ID."""
        self._project_id = project_id

    @property
    def client(self) -> Client:
        """Get the client."""
        return self._client



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   api.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   API class for the inference server.
"""

# pylint: disable=R0902

import os
from typing import Any, Dict

import litserve as ls
import msgspec

from datature.nexus.cli.deploy.local.consts import RUNTIME_OPTIONS
from datature.nexus.cli.deploy.local.loaders.onnx_loader import ONNXLoader
from datature.nexus.cli.deploy.local.predictors.predictor_factory import (
    PredictorFactory,
)
from datature.nexus.cli.deploy.local.types.prediction_request import PredictRequest
from datature.nexus.cli.deploy.local.types.prediction_response import (
    APIDeploymentPredictionResponse,
    PredictionResponse,
)
from datature.nexus.cli.deploy.local.utils.entropy import (
    calculate_avg_entropy,
    calculate_avg_entropy_per_class,
)
from datature.nexus.cli.deploy.local.utils.image_decoder import (
    decode_image_from_base64,
    decode_image_from_bytes,
    decode_image_from_path,
)
from datature.nexus.cli.deploy.local.utils.management_api import ManagementAPIClient
from datature.nexus.cli.deploy.local.utils.prediction_formatter import (
    PredictionFormatter,
)


class InferenceAPI(ls.LitAPI):
    """Inference API class.

    This class is used to create a local deployment server.
    """

    def __init__(
        self,
        secret_key: str,
        project_id: str,
        artifact_id: str,
        runtime: str,
    ):
        super().__init__()

        self.secret_key = secret_key
        self.project_id = project_id
        self.artifact_id = artifact_id
        self.runtime = runtime

        if runtime not in RUNTIME_OPTIONS:
            raise ValueError(f"Unsupported runtime: {runtime}")

        self.execution_provider = RUNTIME_OPTIONS[runtime]
        self._management_api_client = ManagementAPIClient(
            secret_key=self.secret_key,
            project_id=self.project_id,
        )

        self.model = None
        self.predictor = None
        self.formatter = PredictionFormatter()

        self.current_request = None

    def health(self, **kwargs) -> bool:  # pylint: disable=unused-argument
        """Check the health of the server."""
        return True

    def setup(self, device: Any):  # pylint: disable=unused-argument
        """Initialize the model and set up the inference environment."""
        self.model = ONNXLoader(
            artifact_id=self.artifact_id,
            execution_provider=self.execution_provider,
            model_format="ONNX",
            management_api_client=self._management_api_client,
        ).load()

        self.predictor = PredictorFactory.create(
            backend=self.model.backend,
            task_type=self.model.task_type,
            model_type=self.model.model_type,
            model=self.model,
        )

    def decode_request(
        self, request: dict, **kwargs  # pylint: disable=unused-argument
    ) -> Dict[str, Any]:
        """Decode the incoming request to get the image data.

        Returns:
            Dictionary containing the image data and other metadata, if applicable.
        """
        self.current_request = msgspec.convert(request, type=PredictRequest)

        if isinstance(self.current_request.data, str):
            if self.current_request.encoding == "base64":
                return decode_image_from_base64(
                    self.current_request.data, self.current_request.response_format
                )

            if os.path.exists(self.current_request.data):
                return decode_image_from_path(self.current_request.data)

        if isinstance(self.current_request.data, bytes):
            return decode_image_from_bytes(
                self.current_request.data, self.current_request.response_format
            )

        raise ValueError("Invalid image data")

    def predict(self, x: Dict[str, Any], **kwargs):  # pylint: disable=unused-argument
        """Run inference on the input image."""
        return {
            "predictions": self.predictor.run(**x),
            **x,
        }

    def encode_response(
        self, output: Dict[str, Any], **kwargs  # pylint: disable=unused-argument
    ) -> dict:
        """Encode the model output into the response format."""
        formatted_predictions = self.formatter.run(
            task_type=self.model.task_type,
            response_format=self.current_request.response_format,
            label_map=self.model.label_map,
            skeletons=(
                self.model.skeletons if hasattr(self.model, "skeletons") else None
            ),
            **output,
        )

        avg_entropy = calculate_avg_entropy(formatted_predictions)
        avg_entropy_for_class = calculate_avg_entropy_per_class(
            formatted_predictions, self.model.label_map
        )

        prediction_response = PredictionResponse(
            predictions=formatted_predictions,
            avg_entropy=avg_entropy,
            avg_entropy_for_class=avg_entropy_for_class,
            image=None,
            warnings=None,
        )

        response = APIDeploymentPredictionResponse(
            tags=self.model.label_map.to_dict(),
            response=prediction_response,
        )

        return msgspec.to_builtins(response)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   predictor_factory.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Factory class for creating model predictors.
"""

from typing import Dict, Tuple, Type

from datature.nexus.cli.deploy.local.predictors.base_predictor import BasePredictor
from datature.nexus.cli.deploy.local.predictors.onnx_predictor import (
    ONNXClassificationPredictor,
    ONNXObjectDetectionPredictor,
    ONNXSegmentationPredictor,
    ONNXSemantic3DSegmentationPredictor,
    ONNXUltralyticsPredictor,
)
from datature.nexus.cli.deploy.local.utils.inference import PredictionTaskType
from datature.nexus.cli.deploy.local.utils.model import ModelBackend, ModelType


class PredictorFactory:
    """Factory class for creating model predictors."""

    _predictors: Dict[
        Tuple[ModelBackend, PredictionTaskType, ModelType], Type[BasePredictor]
    ] = {}

    @classmethod
    def register(
        cls,
        backend: ModelBackend,
        task_type: PredictionTaskType,
        model_type: ModelType,
        predictor_class: Type[BasePredictor],
    ) -> None:
        """Register a predictor class."""
        cls._predictors[(backend, task_type, model_type)] = predictor_class

    @classmethod
    def create(
        cls,
        backend: ModelBackend,
        task_type: PredictionTaskType,
        model_type: ModelType,
        **kwargs,
    ) -> BasePredictor:
        """Create a predictor instance."""
        key = (backend, task_type, model_type)
        if key not in cls._predictors:
            raise ValueError(f"No predictor registered for {key}")

        predictor_class = cls._predictors[key]
        return predictor_class(**kwargs)


# Register default predictors
PredictorFactory.register(
    ModelBackend.ONNX,
    PredictionTaskType.OBJECT_DETECTION,
    ModelType.ULTRALYTICS,
    ONNXUltralyticsPredictor,
)

PredictorFactory.register(
    ModelBackend.ONNX,
    PredictionTaskType.INSTANCE_SEGMENTATION,
    ModelType.ULTRALYTICS,
    ONNXUltralyticsPredictor,
)

PredictorFactory.register(
    ModelBackend.ONNX,
    PredictionTaskType.KEYPOINT_DETECTION,
    ModelType.ULTRALYTICS,
    ONNXUltralyticsPredictor,
)

PredictorFactory.register(
    ModelBackend.ONNX,
    PredictionTaskType.IMAGE_CLASSIFICATION,
    ModelType.ULTRALYTICS,
    ONNXUltralyticsPredictor,
)

PredictorFactory.register(
    ModelBackend.ONNX,
    PredictionTaskType.OBJECT_DETECTION,
    ModelType.OBJECT_DETECTION,
    ONNXObjectDetectionPredictor,
)

PredictorFactory.register(
    ModelBackend.ONNX,
    PredictionTaskType.INSTANCE_SEGMENTATION,
    ModelType.OBJECT_DETECTION,
    ONNXSegmentationPredictor,
)

PredictorFactory.register(
    ModelBackend.ONNX,
    PredictionTaskType.SEMANTIC_SEGMENTATION,
    ModelType.OBJECT_DETECTION,
    ONNXSegmentationPredictor,
)

PredictorFactory.register(
    ModelBackend.ONNX,
    PredictionTaskType.IMAGE_CLASSIFICATION,
    ModelType.POLYBIUS_V1,
    ONNXClassificationPredictor,
)

PredictorFactory.register(
    ModelBackend.ONNX,
    PredictionTaskType.OBJECT_DETECTION,
    ModelType.POLYBIUS_V1,
    ONNXObjectDetectionPredictor,
)

PredictorFactory.register(
    ModelBackend.ONNX,
    PredictionTaskType.INSTANCE_SEGMENTATION,
    ModelType.POLYBIUS_V1,
    ONNXSegmentationPredictor,
)

PredictorFactory.register(
    ModelBackend.ONNX,
    PredictionTaskType.SEMANTIC_SEGMENTATION,
    ModelType.POLYBIUS_V1,
    ONNXSegmentationPredictor,
)

PredictorFactory.register(
    ModelBackend.ONNX,
    PredictionTaskType.SEMANTIC_3D_SEGMENTATION,
    ModelType.POLYBIUS_V1,
    ONNXSemantic3DSegmentationPredictor,
)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   asset.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Asset API
"""

from typing import Any, Dict, List, Optional, Union
from urllib.parse import quote, unquote

from datature.nexus import models
from datature.nexus.api.asset.upload_session import UploadSession
from datature.nexus.api.operation import Operation
from datature.nexus.api.types import (
    AssetCustomMetadata,
    AssetCustomMetadataBatch,
    AssetFilenameFrameParent,
    AssetFilenameParent,
    AssetFilter,
    AssetFrameParent,
    AssetMetadata,
    AssetParent,
    Pagination,
)
from datature.nexus.client_context import ClientContext, RestContext


class Asset(RestContext):
    """Datature Annotation API Resource."""

    def __init__(self, client_context: ClientContext):
        """Initialize the API Resource."""
        super().__init__(client_context)
        self.operation = Operation(client_context)

    def list(
        self,
        pagination: Union[Pagination, dict, None] = None,
        filters: Union[AssetFilter, dict, None] = None,
    ) -> models.PaginationResponse[models.Asset]:
        """Retrieves a list of all assets in the project.

        :param pagination: A dictionary containing the limit of
                            the number of assets to be returned in each page (defaults to 100),
                            and the page cursor for page selection (defaults to the first page).
        :param filters: A dictionary containing the filters of
                            the assets to be returned.
        :return: A msgspec struct of pagination response with the following structure:

                .. code-block:: python

                    PaginationResponse(
                        next_page='T2YAGDY1NWFlNDcyMzZkiMDYwMTQ5N2U2',
                        prev_page=None,
                        data=[
                            Asset(
                                id='asset_8208740a-2d9c-46e8-abb9-5777371bdcd3',
                                filename='boat180.png',
                                project='proj_cd067221d5a6e4007ccbb4afb5966535',
                                status='None',
                                create_date=1701927649302,
                                url='',
                                metadata=AssetMetadata(
                                    file_size=186497,
                                    mime_type='image/png',
                                    height=243,
                                    width=400,
                                    groups=['main'],
                                    custom_metadata={'captureAt': '2021-03-10T09:00:00Z'}
                                ),
                                statistic=AssetAnnotationsStatistic(
                                    tags_count=[],
                                    total_annotations=0
                                )
                            )
                        ]
                    )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.assets.list({
                            "limit": 2,
                            "page": "ZjYzYmJkM2FjN2UxOTA4ZmU0ZjE0Yjk5Mg"
                        }, filters={
                            "status": "Annotated",
                            "groups": ["main"],
                        })

                        # or
                        project.assets.list(
                            nexus.ApiTypes.Pagination(
                                limit= 2,
                                page="ZjYzYmJkM2FjN2UxOTA4ZmU0ZjE0Yjk5Mg"
                            ),
                            filters=nexus.ApiTypes.AssetFilter(status="Annotated", groups=["main"])
                        )

        """
        assert isinstance(pagination, (Pagination, dict, type(None)))
        assert isinstance(filters, (AssetFilter, dict, type(None)))

        if isinstance(pagination, dict):
            pagination = Pagination(**pagination)
        if pagination is None:
            pagination = Pagination()

        if isinstance(filters, dict):
            filters = AssetFilter(**filters)
        if filters is None:
            filters = AssetFilter()

        return self.requester.GET(
            f"/projects/{self.project_id}/assets",
            query={**pagination.to_json(), **filters.to_json()},
            response_type=models.PaginationResponse[models.Asset],
        )

    def get(self, asset_id_or_name: str) -> models.Asset:
        """Retrieves a specific asset using the asset ID or file name.

        :param asset_id_or_name: The ID or file name of the asset as a string.
        :return: A msgspec struct containing the metadata of one asset with the following structure:

                .. code-block:: python

                        Asset(
                            id='asset_8208740a-2d9c-46e8-abb9-5777371bdcd3',
                            filename='boat180.png',
                            project='proj_cd067221d5a6e4007ccbb4afb5966535',
                            status='None',
                            create_date=1701927649302,
                            url='',
                            metadata=AssetMetadata(
                                file_size=186497,
                                mime_type='image/png',
                                height=243,
                                width=400,
                                groups=['main'],
                                custom_metadata={'captureAt': '2021-03-10T09:00:00Z'}
                            ),
                            statistic=AssetAnnotationsStatistic(
                                tags_count=[],
                                total_annotations=0
                            )
                        )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")
                        project.assets.get("asset_6aea3395-9a72-4bb5-9ee0-19248c903c56")
        """
        assert isinstance(asset_id_or_name, str)
        return self.requester.GET(
            f"/projects/{self.project_id}/assets/{asset_id_or_name}",
            response_type=models.Asset,
        )

    def update(
        self, asset_id_or_name: str, asset_meta: Union[AssetMetadata, dict]
    ) -> models.Asset:
        """Updates the metadata of a specific asset.

        :param asset_id_or_name: The ID or file name of the asset as a string.
        :param asset_meta: The new metadata of the asset to be updated.
        :return: A msgspec struct containing the metadata of one asset with the following structure:

                .. code-block:: python

                        Asset(
                            id='asset_f4dcb429-0332-4dd6-a1b4-fee794031ba6',
                            project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                            filename='boat194.png',
                            status='None',
                            create_date=1701927649302,
                            url='',
                            metadata=AssetMetadata(
                                file_size=172676,
                                mime_type='image/png',
                                height=384,
                                width=422,
                                groups=['main'],
                                custom_metadata={'captureAt': '2021-03-10T09:00:00Z'}
                            ),
                            statistic=AssetAnnotationsStatistic(
                                tags_count=[],
                                total_annotations=0
                            )
                        )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.assets.update(
                            "asset_6aea3395-9a72-4bb5-9ee0-19248c903c56",
                            {
                                "status": "Annotated"
                            }
                        )
        """
        assert isinstance(asset_id_or_name, str)
        assert isinstance(asset_meta, (AssetMetadata, dict))

        if isinstance(asset_meta, dict):
            asset_meta = AssetMetadata(**asset_meta)

        return self.requester.PATCH(
            f"/projects/{self.project_id}/assets/{asset_id_or_name}",
            request_body=asset_meta.to_json(),
            response_type=models.Asset,
        )

    def delete(self, asset_id_or_name: str) -> models.DeleteResponse:
        """Deletes a specific asset from the project.

        :param asset_id_or_name: The ID or file name of the asset as a string.
        :return: A msgspec struct containing the
            deleted asset ID and the deletion status with the following structure.

                .. code-block:: python

                    DeleteResponse(deleted=True, id='asset_8208740a-2d9c-46e8-abb9-5777371bdcd3')

        :example:

                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.assets.delete(
                            "asset_6aea3395-9a72-4bb5-9ee0-19248c903c56",
                        )
        """
        assert isinstance(asset_id_or_name, str)
        return self.requester.DELETE(
            f"/projects/{self.project_id}/assets/{asset_id_or_name}",
            response_type=models.DeleteResponse,
        )

    def create_upload_session(
        self, groups: Optional[List[str]] = None, background: bool = False
    ) -> UploadSession:
        """
        Creates a new upload session with specified
        groups and an option to run in the background.

        This method initializes and returns an UploadSession object,
        which can be used to manage file uploads within the system.

        :param groups: A list of group names to categorize the upload. Default is None.
        :param background: A flag indicating whether
                            the upload should run in the background. Default is False.
        :return: UploadSession: An instance of the UploadSession class.
        :example:

                .. code-block:: python

                    from datature.nexus import Client

                    project = Client("5aa41e8ba........").get_project("proj_b705a........")

                    upload_session = project.assets.create_upload_session(
                                                groups=["main"],
                                                background=True
                                            )

                    with upload_session as session:
                        # add assets path to upload session
                        upload_session.add_path(
                            "folder/path/to/files",
                            custom_metadata={"key": "value"}
                        )

                        upload_session.add_bytes(
                            b"bytes/of/file",
                            "filename",
                            custom_metadata={"key": "value"}
                        )
        """
        assert isinstance(groups, (list, type(None)))
        assert isinstance(background, bool)

        return UploadSession(self._context, groups, background)

    def list_groups(self, groups: Union[List[str], None] = None) -> models.AssetGroups:
        """Retrieve asset statistics categorized by asset group and asset status.

        :param groups: A string array of name(s) of asset group(s).
        :return: A list of msgspec struct of
            the categorized asset statistics with the following structure:

                .. code-block:: python

                        [
                            AssetGroup(
                                group='1',
                                statistic=AssetGroupStatistic(
                                    total_assets=1,
                                    annotated_assets=0,
                                    reviewed_assets=0,
                                    to_fixed_assets=0,
                                    completed_assets=0
                                )
                            ),
                            AssetGroup(
                                group='main',
                                statistic=AssetGroupStatistic(
                                    total_assets=503,
                                    annotated_assets=0,
                                    reviewed_assets=0,
                                    to_fixed_assets=0,
                                    completed_assets=0
                                )
                            )
                        ]

        :example:

                .. code-block:: python

                    from datature.nexus import Client

                    project = Client("5aa41e8ba........").get_project("proj_b705a........")
                    project.assets.list_groups()
        """
        assert isinstance(groups, (list, type(None)))

        return self.requester.GET(
            f"/projects/{self.project_id}/assetgroups",
            query={"group": groups},
            response_type=models.AssetGroups,
        )

    def get_custom_metadata(self, asset_id_or_name: str) -> models.AssetCustomMetadatas:
        """Retrieve custom metadata for a specific asset.

        This method fetches all custom metadata associated with a given asset,
        including metadata for specific frames if the asset is a video.

        :param asset_id_or_name: The ID or file name of the asset as a string.
        :return: A msgspec struct containing the custom metadata of the asset with the following structure:

                .. code-block:: python

                        AssetCustomMetadatas(
                            custom_metadatas=[
                                AssetCustomMetadata(
                                    custom_metadata={"captureAt": "2021-03-10T09:00:00Z"}
                                )
                            ]
                        )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")
                        project.assets.get_custom_metadata("asset_6aea3395-9a72-4bb5-9ee0-19248c903c56")
        """
        assert isinstance(asset_id_or_name, str)

        return self.requester.GET(
            f"/projects/{self.project_id}/assets/{asset_id_or_name}/customMetadatas",
            response_type=models.AssetCustomMetadatas,
        )

    def add_custom_metadata(
        self,
        asset_id_or_name: str,
        custom_metadata: Dict[str, Union[str, int, float, bool]],
        frame: Optional[int] = None,
    ) -> models.AssetCustomMetadata:
        """Add custom metadata to a specific asset.

        This method allows you to add custom metadata to an asset. For video assets,
        you can must specify the frame number.

        :param asset_id_or_name: The ID or file name of the asset as a string.
        :param custom_metadata: A dictionary of custom metadata to add to the asset.
                               Supported value types are string, int, float, and bool.
        :param frame: Required frame number for video assets. Optional for image assets.
        :return: A msgspec struct containing the added custom metadata with the following structure:

                .. code-block:: python

                        AssetCustomMetadata(
                            custom_metadata={"captureAt": "2021-03-10T09:00:00Z"}
                        )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        # Add metadata to asset level
                        project.assets.add_custom_metadata(
                            "asset_6aea3395-9a72-4bb5-9ee0-19248c903c56",
                            {"captureAt": "2021-03-10T09:00:00Z"}
                        )

                        # Add metadata to specific frame (for video assets)
                        project.assets.add_custom_metadata(
                            "asset_6aea3395-9a72-4bb5-9ee0-19248c903c56",
                            {"event": "motion_detected"},
                            frame=15
                        )
        """
        assert isinstance(asset_id_or_name, str)
        assert isinstance(custom_metadata, dict)

        if not all(
            isinstance(key, str) and isinstance(value, (str, int, float, bool))
            for key, value in custom_metadata.items()
        ):
            raise ValueError(
                "custom_metadata must be a dictionary with keys of string, and values of string, int, float, or bool"
            )

        if frame:
            parent = AssetFrameParent(frame=frame)
        else:
            parent = AssetParent()

        custom_meta = AssetCustomMetadata(
            parent=parent, custom_metadata=custom_metadata
        )

        return self.requester.PUT(
            f"/projects/{self.project_id}/assets/{asset_id_or_name}/customMetadatas",
            request_body=custom_meta.to_json(),
            response_type=models.AssetCustomMetadata,
        )

    def delete_custom_metadata(
        self, asset_id_or_name: str, custom_metadata_id: str
    ) -> models.DeleteResponse:
        """Delete custom metadata from a specific asset.

        This method removes a specific custom metadata entry from an asset.
        The custom_metadata_id can be obtained from the get_custom_metadata method.

        :param asset_id_or_name: The ID or file name of the asset as a string.
        :param custom_metadata_id: The ID of the custom metadata to delete.
        :return: A msgspec struct containing the deletion status with the following structure:

                .. code-block:: python

                        DeleteResponse(
                            deleted=True,
                            id='custommeta_customMetadata#assetParent:59d4e65e05e7360235526f75698ef3e0:97a83640-f56d-440c-9095-362de2d12681'
                        )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")
                        project.assets.delete_custom_metadata(
                            "asset_6aea3395-9a72-4bb5-9ee0-19248c903c56",
                            "custommeta_customMetadata#assetParent:59d4e65e05e7360235526f75698ef3e0:97a83640-f56d-440c-9095-362de2d12681",
                        )
        """
        assert isinstance(asset_id_or_name, str)
        assert isinstance(custom_metadata_id, str)

        # Ensures proper encoding no matter the encoding state
        encoded_custom_metadata_id = quote(unquote(custom_metadata_id))

        return self.requester.DELETE(
            f"/projects/{self.project_id}/assets/{asset_id_or_name}"
            f"/customMetadatas/{encoded_custom_metadata_id}",
            response_type=models.DeleteResponse,
        )

    def add_custom_metadata_batch(
        self,
        custom_metadata_batch: List[Dict[str, Dict[str, Any]]],
    ) -> models.AssetCustomMetadata:
        """Add custom metadata to multiple assets in a single operation.

        This method allows you to add custom metadata to multiple assets or frames
        in a single batch operation, which is more efficient than individual calls.
        Each item in the batch should specify the target asset (by ID or filename)
        and a frame number for video assets.

        :param custom_metadata_batch: A list of dictionaries, where each dictionary contains:
                                     - "parent": A dictionary with either "asset_id" or "filename",
                                       and "frame" for video assets
                                     - "custom_metadata": A dictionary of metadata key-value pairs
        :return: A msgspec struct containing the batch operation response.
        :raises ValueError: If the parent specification is invalid (missing asset_id or filename).

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        # Batch add metadata to multiple assets
                        batch_data = [
                            {
                                "parent": {"asset_id": "asset_123", "frame": 5},
                                "custom_metadata": {"event": "motion", "confidence": 0.95}
                            },
                            {
                                "parent": {"filename": "image.jpg"},
                                "custom_metadata": {"location": "camera_1", "timestamp": "2021-03-10T09:00:00Z"}
                            }
                        ]

                        project.assets.add_custom_metadata_batch(batch_data)
        """

        assert isinstance(custom_metadata_batch, list)

        custom_metas = AssetCustomMetadataBatch(custom_metadata_batch=[])

        for item in custom_metadata_batch:
            parent = item.get("parent", {})

            asset_id = parent.get("asset_id")
            filename = parent.get("filename")
            frame = parent.get("frame")

            if not all(
                isinstance(key, str) and isinstance(value, (str, int, float, bool))
                for key, value in item.get("custom_metadata", {}).items()
            ):
                raise ValueError(
                    "custom_metadata must be a dictionary with keys of string, and values of string, int, float, or bool"
                )

            if asset_id and frame:
                custom_metas.custom_metadata_batch.append(
                    AssetCustomMetadata(
                        parent=AssetFrameParent(asset_id=asset_id, frame=frame),
                        custom_metadata=item.get("custom_metadata", {}),
                    )
                )

            elif asset_id and not frame:
                custom_metas.custom_metadata_batch.append(
                    AssetCustomMetadata(
                        parent=AssetParent(asset_id=asset_id),
                        custom_metadata=item.get("custom_metadata", {}),
                    )
                )

            elif filename and frame:
                custom_metas.custom_metadata_batch.append(
                    AssetCustomMetadata(
                        parent=AssetFilenameFrameParent(filename=filename, frame=frame),
                        custom_metadata=item.get("custom_metadata", {}),
                    )
                )

            elif filename and not frame:
                custom_metas.custom_metadata_batch.append(
                    AssetCustomMetadata(
                        parent=AssetFilenameParent(filename=filename),
                        custom_metadata=item.get("custom_metadata", {}),
                    )
                )

            else:
                raise ValueError(
                    "Invalid parent. Parent must contain either asset_id or filename."
                )

        return self.requester.POST(
            f"/projects/{self.project_id}/customMetadataBulkUpdates",
            request_body=custom_metas.to_json(),
            response_type=models.AssetCustomMetadataBatch,
        )

    def delete_custom_metadata_batch(
        self,
        custom_metadata_ids: List[str],
    ) -> models.DeleteCustomMetadataBatchResponse:
        """Delete custom metadata from multiple assets in a single operation.

        This method allows you to delete multiple custom metadata entries across
        different assets in a single batch operation. The custom_metadata_ids can
        be obtained from the get_custom_metadata method for each asset.

        :param custom_metadata_ids: A list of custom metadata IDs to delete.
        :return: A msgspec struct containing the batch deletion response.
                 If the list is empty, returns an empty response.

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        # Delete multiple custom metadata entries
                        metadata_ids = [
                            "custommeta_customMetadata#assetParent:59d4e65e05e7360235526f75698ef3e0:97a83640-f56d-440c-9095-362de2d12681",
                            "custommeta_customMetadata#assetParent:59d4e65e05e7360235526f75698ef3e0:a0d7813c-47c5-4da0-a6e7-4b73f5539cf5",
                        ]
                        project.assets.delete_custom_metadata_batch(metadata_ids)
        """

        assert isinstance(custom_metadata_ids, list)

        if len(custom_metadata_ids) == 0:
            return models.DeleteCustomMetadataBatchResponse()

        return self.requester.POST(
            f"/projects/{self.project_id}/customMetadataBulkUpdates/delete",
            request_body={"ids": custom_metadata_ids},
            response_type=models.DeleteCustomMetadataBatchResponse,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   functions.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Outpost user functions.
"""

# pylint: disable=W0718,E1120

import logging
import re
import sys

import pytz
from InquirerPy import inquirer
from InquirerPy.validator import ValidationError, Validator

import datature.nexus.cli.consts as CONSTS
import datature.nexus.cli.outpost.consts as OUTPOST_CONSTS
from datature import nexus
from datature.nexus.cli import messages
from datature.nexus.cli.config import Config
from datature.nexus.cli.outpost.common import log_errors
from datature.nexus.cli.outpost.consts import (
    OUTPOST_CONFIG_ROOT_DIR,
    RESERVED_DEVICE_TAGS,
)
from datature.nexus.cli.outpost.manager import OutpostManager
from datature.nexus.cli.outpost.types import OutpostDeviceConfig
from datature.nexus.cli.outpost.utils import check_system_specs, validate_device
from datature.nexus.error import Error

TIMEZONE_RE = r"[A-Za-z]+/[A-Za-z_]+"


class UniqueNameAndIdValidator(Validator):
    """Validator to check if the name is unique."""

    def __init__(self, outpost_device_config: OutpostDeviceConfig):
        self.outpost_device_config = outpost_device_config

    def validate(self, document) -> None:
        """Validate the input.

        :param document: Input document.
        """
        display_name = document.text.strip()
        if len(display_name) <= 0:
            raise ValidationError(
                message="Name cannot be empty.",
                cursor_position=document.cursor_position,
            )

        validation_errors = validate_device(
            self.outpost_device_config, display_name
        ).get("validationErrors", [])

        if "DisplayNameConflict" in validation_errors:
            raise ValidationError(
                message="A registered device with the same name has been found. Please enter a unique name.",
                cursor_position=document.cursor_position,
            )


def prompt_outpost_details(outpost_device_config: OutpostDeviceConfig):
    """Prompt for Outpost device details.

    :param outpost_device_config (OutpostDeviceConfig): Outpost device configuration.
    """
    print()

    outpost_device_config.name = (
        inquirer.text(
            message="Enter a descriptive name for this device:",
            mandatory=True,
            validate=UniqueNameAndIdValidator(outpost_device_config),
        )
        .execute()
        .strip()
    )

    timezone_choices = [tz for tz in pytz.all_timezones if re.match(TIMEZONE_RE, tz)]
    timezone = inquirer.fuzzy(
        message="Select your preferred timezone for metrics and logs display:",
        choices=timezone_choices,
        match_exact=True,
        default="UTC",
        exact_symbol="",
        max_height=CONSTS.INQUIRER_CHOICES_MAX_VISIBLE_ENTRIES,
        border=True,
    ).execute()

    outpost_device_config.tags["timezone"] = str(pytz.timezone(timezone).zone)

    tags = (
        inquirer.text(
            message="Enter device tags as name:value pairs, separated by commas (optional):",
            default="",
            long_instruction=(
                "Tags are key-value pairs that can be used to filter and group devices for "
                "fleet management. For example, you can tag devices with "
                "location:warehouse-a, floor:2, etc."
            ),
            validate=lambda result: all(
                tag.split(":")[0].strip() not in RESERVED_DEVICE_TAGS
                for tag in result.split(",")
            ),
            invalid_message=(
                "Please ensure that your tag list does not contain any of the reserved tags:"
                f" {', '.join(RESERVED_DEVICE_TAGS)}"
            ),
        )
        .execute()
        .split(",")
    )

    for tag in tags:
        if ":" in tag:
            tag_name, tag_value = tag.split(":")
            outpost_device_config.tags[tag_name.strip()] = tag_value.strip()

    outpost_device_config.write_to_file()


def create_configuration():
    """Create new Outpost configuration."""
    OUTPOST_CONFIG_ROOT_DIR.mkdir(parents=True, exist_ok=True)

    try:
        outpost_device_config = OutpostDeviceConfig.read_from_file()

        logging.basicConfig(level=logging.CRITICAL)
        cli_config = Config()
        outpost_device_config.device_credentials.secret_key = (
            cli_config.get_default_project().get("project_secret", "")
        )
        outpost_device_config.device_credentials.default_project_id = (
            cli_config.get_default_project().get("project_id", "")
        )
        outpost_device_config.device_credentials.workspace_id = (
            nexus.Client(outpost_device_config.device_credentials.secret_key)
            .get_info()
            .id.strip("ws_")
        )

        logging.basicConfig(level=logging.INFO)
    except Error as exc:
        if "Workspace ID / Secret Key combination" in exc.message:
            outpost_device_config.device_credentials.secret_key = ""
            outpost_device_config.device_credentials.workspace_id = ""
        elif "No access rights" in exc.message:
            print(messages.AUTHENTICATION_FAILED_MESSAGE)
            sys.exit(1)
        else:
            print(exc)
            sys.exit(1)

    outpost_device_config.write_to_file()

    OutpostManager().create_configuration()


def install_outpost():
    """Install Outpost on device."""
    OUTPOST_CONFIG_ROOT_DIR.mkdir(parents=True, exist_ok=True)

    if not OUTPOST_CONSTS.OUTPOST_CONFIG_FILE_PATH.exists():
        OUTPOST_CONSTS.OUTPOST_CONFIG_FILE_PATH.touch()

    try:
        outpost_device_config = OutpostDeviceConfig.read_from_file()

        logging.basicConfig(level=logging.CRITICAL)
        cli_config = Config()
        outpost_device_config.device_credentials.secret_key = (
            cli_config.get_default_project().get("project_secret", "")
        )
        outpost_device_config.device_credentials.default_project_id = (
            cli_config.get_default_project().get("project_id", "").strip("proj_")
        )
        outpost_device_config.device_credentials.workspace_id = (
            nexus.Client(outpost_device_config.device_credentials.secret_key)
            .get_info()
            .id.strip("ws_")
        )

        logging.basicConfig(level=logging.INFO)
    except Error as exc:
        if "Workspace ID / Secret Key combination" in exc.message:
            outpost_device_config.device_credentials.secret_key = ""
            outpost_device_config.device_credentials.workspace_id = ""
        elif "No access rights" in exc.message:
            print(messages.AUTHENTICATION_FAILED_MESSAGE)
            sys.exit(1)
        else:
            print(exc)
            sys.exit(1)

    try:
        check_system_specs(outpost_device_config)

        print()
        license_agreed = inquirer.confirm(
            message=(
                "Please confirm that you have reviewed and accepted the "
                "Datature Outpost License Agreement.\n\n"
                "By proceeding with installation, you acknowledge compliance "
                "with the terms and conditions.\n"
                "Full agreement available at: https://datature.io/legal/outpost/license.txt"
            ),
            default=False,
        ).execute()

        if not license_agreed:
            sys.exit(0)

        prompt_outpost_details(outpost_device_config)
    except KeyboardInterrupt:
        sys.exit(0)
    except Error as exc:
        print(exc)
        sys.exit(1)
    except Exception as exc:
        print(exc)
        sys.exit(1)

    OutpostManager().install()


def uninstall_outpost():
    """Uninstall Outpost from device."""
    try:
        outpost_device_config = OutpostDeviceConfig.read_from_file()
        if not outpost_device_config.id:
            print(messages.NO_OUTPOST_DEVICE_REGISTERED_MESSAGE)
            sys.exit(1)

        continue_uninstall = inquirer.confirm(
            message="Please confirm that you want to uninstall Datature Outpost from this device:",
            default=False,
        ).execute()

        if not continue_uninstall:
            sys.exit(0)

        OutpostManager().uninstall()
    except KeyboardInterrupt:
        sys.exit(0)
    except (Error, Exception):
        log_errors()
        sys.exit(1)


def pause_outpost():
    """Pause Outpost manually on device."""
    try:
        outpost_device_config = OutpostDeviceConfig.read_from_file()
        if not outpost_device_config.id:
            print(messages.NO_OUTPOST_DEVICE_REGISTERED_MESSAGE)
            sys.exit(1)

        OutpostManager().pause()

    except KeyboardInterrupt:
        sys.exit(1)
    except (Error, Exception):
        log_errors()
        sys.exit(1)


def resume_outpost():
    """Resume Outpost manually on device."""
    try:
        outpost_device_config = OutpostDeviceConfig.read_from_file()
        if not outpost_device_config.id:
            print(messages.NO_OUTPOST_DEVICE_REGISTERED_MESSAGE)
            sys.exit(1)

        OutpostManager().resume()

    except KeyboardInterrupt:
        sys.exit(1)
    except (Error, Exception):
        log_errors()
        sys.exit(1)


def display_outpost_status():
    """Displays Outpost device status."""

    try:
        outpost_device_config = OutpostDeviceConfig.read_from_file()
        if not outpost_device_config.id:
            print(messages.NO_OUTPOST_DEVICE_REGISTERED_MESSAGE)
            sys.exit(1)

        OutpostManager().status()

    except KeyboardInterrupt:
        sys.exit(1)
    except (Error, Exception):
        log_errors()
        sys.exit(1)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   messages.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Outpost information messages.
"""

from datature.nexus.cli.outpost.consts import DOCS_ERROR_HANDLING_URL, SUPPORT_EMAIL

INSTALL_PIP_PACKAGE_ERROR_MESSAGE = (
    "Failed to install {package_name}.\n"
    "You may need to install manually by running `pip install {package_name}` in your environment.\n"
    "Also ensure that your pip version is up-to-date by running `pip install --upgrade pip`."
)

OUTPOST_UNKNOWN_ERROR_MESSAGE = (
    "Error: An unknown error occurred while processing Datature Outpost. Full traceback logged to {}.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
OUTPOST_TRACEBACK_MESSAGE = "Full traceback logged to {}."


SERVER_CONNECTION_ERROR_MESSAGE = (
    "Error: Failed to connect to server. Ensure you have an active internet connection.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
FEATURE_NOT_AVAILABLE_ERROR_MESSAGE = (
    "Error: Your workspace does not have access to this feature.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
INSUFFICIENT_DEVICE_QUOTA_ERROR_MESSAGE = (
    "Error: Your workspace has reached the maximum number of devices.\n"
    "Please remove any unused devices from your workspace.\n"
    f"For more information on quotas, visit: {DOCS_ERROR_HANDLING_URL} "
    f"or contact support if you wish to increase your quotas: {SUPPORT_EMAIL}"
)
SERVER_FORBIDDEN_ERROR_MESSAGE = (
    "Error: Failed to ping server. "
    "The Workspace ID / Secret Key combination has been invalidated. "
    "A user in your workspace may have generated a new key.\n"
    "Please obtain the new key, or regenerate one for your workspace on Nexus: "
    "https://nexus.datature.io/workspace/{}/settings#key\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)
SERVER_UNKNOWN_ERROR_MESSAGE = (
    "{}\n"
    "Error: Failed to ping server.\n"
    f"For more details, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)

MISSING_CONFIGURATION_FILES_ERROR_MESSAGE = (
    "Error: Missing configuration files detected. Your provided folder should contain the following files: {}\n"
    "Please run `datature outpost configure` to create the configuration files.\n"
    f"For more information, visit: {DOCS_ERROR_HANDLING_URL} or contact support: {SUPPORT_EMAIL}"
)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   requester.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   A wrapper to make HTTP requests
"""

import json
import logging
import platform
from typing import List, Optional, Tuple
from urllib.parse import urljoin

import msgspec
from requests import Response, Session

from datature.nexus import config, error
from datature.nexus.version import __version__

system = platform.system()
machine = platform.machine()
python_version = platform.python_version()

logger = logging.getLogger("datature-nexus")

# pylint: disable=C0103,R0913,R0914,R0917


class Requester:
    """An HTTP requester."""

    _timeout: int
    _base_url: str
    _session: Session
    _cert: Optional[Tuple[str, str]]
    secret_key: Optional[str]

    def __init__(
        self,
        secret_key: Optional[str],
        mtls_certificate_file_path: Optional[str],
        mtls_private_key_file_path: Optional[str],
        endpoint: str,
    ):
        """Create an HTTP requester.

        :param secret_key: The secret key
        :param mtls_certificate_file_path: The path to the mTLS certificate file
        :param mtls_private_key_file_path: The path to the mTLS private key file
        :param endpoint: The base URL for all requests
        """
        self._timeout = config.REQUEST_TIME_OUT_SECONDS
        self._base_url = endpoint

        self._session = Session()

        self._cert = (
            (mtls_certificate_file_path, mtls_private_key_file_path)
            if mtls_certificate_file_path and mtls_private_key_file_path
            else None
        )

        self.secret_key = secret_key

        # Check Secret
        if not secret_key and not self._cert:
            raise AttributeError(
                "Either the secret key or mTLS certificate files are required"
            )

    def _request(
        self,
        method,
        path,
        response_type,
        query=None,
        request_body=None,
        request_headers=None,
        request_params=None,
        ignore_errno=None,
    ):
        """Create an HTTP requester.

        :param method: The request method
        :param path: The request path
        :param query: The query object
        :param request_body: The request body
        :param request_headers: If have custom request headers
        :param request_params: The request parameters
        :param ignore_errno: The error codes to ignore
        """
        method = method.upper()
        assert method in ["GET", "HEAD", "DELETE", "POST", "PUT", "PATCH", "OPTIONS"]

        # Assemble queries and request body.
        post_data = None
        post_files = None

        absolute_url = urljoin(self._base_url, path)

        if query:
            encoded_params = self._make_query_string(query)
            absolute_url = f"{absolute_url}?{encoded_params}"

        if request_body:
            post_data = json.dumps(request_body)

        logger.debug("API request: %s %s Body %s", method, absolute_url, post_data)

        if self._cert:
            response = self._session.request(
                method,
                absolute_url,
                headers={
                    "Content-Type": "application/json",
                    "User-Agent": f"DatatureSDK/{__version__} "
                    f"(Python/{python_version}; {system}/{machine})",
                },
                cert=self._cert,
                data=post_data,
                files=post_files,
                params=request_params,
                verify=True,
                timeout=self._timeout,
            )
        else:
            # Assemble request headers
            headers = self._make_headers(method, request_headers)

            # Call request to do a real HTTP call, default timeout 120s
            response = self._session.request(
                method,
                absolute_url,
                headers=headers,
                cert=None,
                data=post_data,
                files=post_files,
                params=request_params,
                timeout=self._timeout,
            )

        self._interpret_response(response, ignore_errno)

        logger.debug(
            "API response: %s %s Body %s", method, absolute_url, response.json()
        )

        return msgspec.json.decode(
            json.dumps(response.json()),
            type=response_type,
        )

    def GET(
        self, path, response_type, query=None, request_params=None, ignore_errno=None
    ):
        """GET Call."""
        return self._request(
            "GET",
            path,
            response_type,
            query=query,
            request_params=request_params,
            ignore_errno=ignore_errno,
        )

    def POST(
        self,
        path,
        response_type,
        query=None,
        request_body=None,
        request_params=None,
        ignore_errno=None,
    ):
        """POST Call."""
        return self._request(
            "POST",
            path,
            response_type,
            query=query,
            request_body=request_body,
            request_params=request_params,
            ignore_errno=ignore_errno,
        )

    def PUT(
        self,
        path,
        response_type,
        query=None,
        request_body=None,
        request_params=None,
        ignore_errno=None,
    ):
        """PUT Call."""
        return self._request(
            "PUT",
            path,
            response_type,
            query=query,
            request_body=request_body,
            request_params=request_params,
            ignore_errno=ignore_errno,
        )

    def PATCH(
        self,
        path,
        response_type,
        query=None,
        request_body=None,
        request_params=None,
        ignore_errno=None,
    ):
        """PATCH Call."""
        return self._request(
            "PATCH",
            path,
            response_type,
            query=query,
            request_body=request_body,
            request_params=request_params,
            ignore_errno=ignore_errno,
        )

    def DELETE(self, path, response_type, ignore_errno=None):
        """DELETE Call."""
        return self._request("DELETE", path, response_type, ignore_errno=ignore_errno)

    def _make_headers(self, method, request_headers):
        """Make request headers

        :param method: The request method
        :param request_headers: The custom request headers
        """
        headers = {
            "Secret-Key": self.secret_key,
            "User-Agent": f"DatatureSDK/{__version__} "
            f"(Python/{python_version}; {system}/{machine})",
        }

        if method in ["POST", "PUT", "PATCH", "GET"]:
            headers["Content-Type"] = "application/json"

        if request_headers is not None:
            for key, value in request_headers.items():
                headers[key] = value

        return headers

    def _make_query_string(self, query):
        """Make query string

        :param query: The query object
        """
        if query is None:
            return ""

        query_string = ""
        for key, value in query.items():
            if isinstance(value, bool):
                query_string += f"{key}={str(value).lower()}&"
            elif isinstance(value, list):
                for item in value:
                    query_string += f"{key}[]={item}&"
            else:
                if value is not None:
                    query_string += f"{key}={value}&"

        return query_string

    def _interpret_response(
        self, response: Response, ignore_errno: Optional[List[int]] = None
    ):
        """Check if need throw Error

        :param response: The request response
        :param ignore_errno: The error code(s) to ignore, if any
        """
        response_code = response.status_code
        response_data = response.json()

        if not 200 <= response_code < 300:
            error_message = response_data["message"]

            if ignore_errno is not None and response_code in ignore_errno:
                logger.debug(
                    "API response errored: %s %s, but ignoring error",
                    response_code,
                    error_message,
                )
                return

            logger.error("API response errored: %s %s", response_code, error_message)

            if response_code == 400:
                raise error.BadRequestError(error_message, response_data)
            if response_code == 403:
                raise error.ForbiddenError(error_message, response_data)
            if response_code == 404:
                raise error.NotFoundError(
                    error_message,
                    response_data,
                )
            if response_code == 429:
                raise error.TooManyRequestsError(error_message, response_data)

            raise error.InternalServerError(error_message, response_data)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   label_map.py
@Author  :   Wei Loon Cheng
@Version :   1.0
@Contact :   hello@datature.io
@License :   Datature Outpost License 1.0
@Desc    :   Label map utilities.
"""

# pylint: disable=C0103

from dataclasses import dataclass
from typing import Dict, Generator, List, Optional, TextIO, Tuple, Union

import google.protobuf.text_format as pb_text_format

from datature.nexus.cli.deploy.local.types.prediction_response import Tag
from datature.nexus.cli.deploy.local.utils.colors import get_color_for_label_id

# pylint: disable=no-name-in-module
from datature.nexus.cli.deploy.local.utils.label_map_pb2 import LabelMap as PBLabelMap


class LabelMapParseError(Exception):
    """Error parsing a label map."""


@dataclass
class LabelMapEntry:
    """Entry in a label map."""

    label_id: int
    name: str
    color: str


class LabelMap:
    """Label map for a model."""

    _by_id: Dict[int, LabelMapEntry]
    _by_name: Dict[str, LabelMapEntry]

    def __init__(self, entries: List[LabelMapEntry]):
        self._by_id = {}
        self._by_name = {}

        for entry in entries:
            if entry.label_id in self._by_id:
                raise LabelMapParseError(
                    f"Tag ID {entry.label_id} already defined as "
                    f"'{self._by_id[entry.label_id].name}'"
                )

            if entry.name in self._by_name:
                raise LabelMapParseError(
                    f"Tag Name {entry.name} already defined as "
                    f"{self._by_name[entry.name].label_id}"
                )

            self._by_id[entry.label_id] = entry
            self._by_name[entry.name] = entry

    @property
    def names(self) -> Dict[str, LabelMapEntry]:
        """Get a dictionary of tag names to LabelMapEntries."""
        return self._by_name

    @property
    def ids(self) -> Dict[int, LabelMapEntry]:
        """Get a dictionary of tag IDs to LabelMapEntries."""
        return self._by_id

    def to_dict(self) -> Dict[str, Tag]:
        """Convert the label map to a dictionary of tag names to Tag objects."""
        return {
            str(entry.label_id): Tag(name=entry.name, color=entry.color)
            for entry in self._by_id.values()
        }

    def __getitem__(self, key: Union[int, str]) -> LabelMapEntry:
        if isinstance(key, int):
            return self._by_id[key]

        if isinstance(key, str):
            return self._by_name[key]

        return None

    def __len__(self) -> int:
        return len(self._by_id)

    def __iter__(self) -> Generator[Tuple[int, str, str], None, None]:
        return (
            (label_id, entry.name, entry.color)
            for label_id, entry in self._by_id.items()
        )


def LabelMapFromPbtxt(
    pbtxt_file_obj: TextIO, color_map: Optional[Dict[str, str]]
) -> LabelMap:
    """Read a label map from a file in PBTXT format."""

    try:
        pbtxt_file_text = pbtxt_file_obj.read()
    except IOError as e:
        raise LabelMapParseError("Unable to read label map from pbtxt") from e

    pb_label_map = PBLabelMap()
    entries: List[LabelMapEntry] = []
    have_background: bool = False
    have_class_zero: bool = False

    try:
        pb_text_format.Parse(pbtxt_file_text, pb_label_map)
    except pb_text_format.ParseError as e:
        raise LabelMapParseError("Could not load label map from pbtxt") from e

    for entry in pb_label_map.item:  # pylint: disable=E1101
        if len(entry.name) < 1:
            raise LabelMapParseError(
                f"Label map entry with ID {entry.id} has an empty name"
            )

        if entry.name.lower() == "background":
            have_background = True

        if entry.id == 0:
            have_class_zero = True

        entries.append(
            LabelMapEntry(
                label_id=entry.id,
                name=entry.name,
                color=(color_map or {}).get(
                    entry.name, get_color_for_label_id(entry.id)
                ),
            )
        )

    if not have_background and not have_class_zero:
        entries.append(LabelMapEntry(label_id=0, name="background", color="#000000"))

    if len(entries) < 1:
        raise ValueError("Label map has no entries")

    return LabelMap(entries)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   init.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   CLI Outpost utils import module.
"""

from .config import (
    create_configuration,
    list_configurations,
    prompt_for_configuration_version,
    prompt_for_runtime_config,
    select_default_artifact,
)
from .device_config import DEVICE_CONFIG
from .helper import interpolate_config_variables, start_spinner
from .system_specs import check_system_specs, get_current_device_info
from .utils import (
    deregister_device,
    format_outpost_runtime_status,
    generate_device_config,
    generate_device_info,
    generate_metadata_spec,
    generate_pem_keys,
    get_certificate_expiry,
    get_device,
    get_outpost_device_config_files,
    get_outpost_runtime_files,
    get_outpost_runtime_status,
    get_runtime_status_message,
    get_service_status,
    install_runtime,
    list_devices,
    parse_systemctl_output,
    patch_device,
    register_device,
    setup_outpost_credentials,
    uninstall_runtime,
    validate_certificate,
    validate_device,
)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   prediction_request.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Prediction request types.
"""

from enum import Enum
from typing import Optional

from msgspec import Struct


class ResponseFormat(Enum):
    """
    Response format enum.
    """

    CLASSIFICATION = "classification"
    RECTANGLE = "rectangle"
    POLYGON = "polygon"
    BITMASK = "bitmask"
    KEYPOINT = "keypoint"
    NIFTI = "nifti"


class PredictRequest(Struct, kw_only=True, rename="camel"):
    """
    Predict request object.

    Attributes:
        project: The project of the request.
        data: The data of the request.
        encoding: The encoding of the request.
        mime_type: The mime type of the request.
    """

    project: Optional[str] = None
    data: Optional[str] = None
    encoding: Optional[str] = None
    mime_type: Optional[str] = None
    response_format: Optional[ResponseFormat] = None



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   config.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   SDK config module
"""

# Default configurations
ASSET_UPLOAD_SESSION_BATCH_SIZE: int = 1000
IMAGE_MAX_SIZE: int = 256 * 1024 * 1024  # 256MB
MEDICAL_3D_MAX_SIZE: int = 1024 * 1024 * 1024  # 1GB
VIDEO_MAX_SIZE: int = 512 * 1024 * 1024 * 1024  # 512GB
ASSET_UPLOAD_SESSION_MULTIPART_MIN_SIZE: int = 1 * 1024 * 1024 * 1024  # 1GB
ASSET_UPLOAD_SESSION_WORKERS_RATIO: float = 0.6
ASSET_UPLOAD_MULTIPART_MAX_RETRIES: int = 3
ANNOTATION_IMPORT_SESSION_MAX_SIZE: int = 100000
ANNOTATION_IMPORT_SESSION_BATCH_SIZE: int = 1000
ANNOTATION_IMPORT_SESSION_BATCH_BYTES: int = 1024 * 1024 * 1024  # 1GB
OPERATION_LOOPING_TIMEOUT_SECONDS: int = 36000
OPERATION_LOOPING_DELAY_SECONDS: int = 8
REQUEST_TIME_OUT_SECONDS = (60, 3600)
FILE_CHUNK_SIZE = 8 * 1024 * 1024  # 8MB



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   functions.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Configuration functions for the local deployment server.
"""

# pylint: disable=W0718,R0913,R0917

import sys
from datetime import datetime
from typing import Dict, Tuple
from uuid import uuid4

import requests
from InquirerPy import inquirer
from InquirerPy.base.control import Choice

from datature.nexus.cli.config import Config
from datature.nexus.cli.deploy.local.config import DeploymentConfig, DeploymentConfigs
from datature.nexus.cli.deploy.local.server import start_server
from datature.nexus.cli.functions import get_default_datature_client
from datature.nexus.config import REQUEST_TIME_OUT_SECONDS


def get_project_credentials() -> Dict[str, str]:
    """Get project credentials from config"""
    config = Config()
    project = config.get_default_project()

    if project is None:
        print("No project found. Please set a default project.")
        sys.exit(1)

    return project


def select_artifact() -> Tuple[str, str]:
    """Prompt user to select an artifact"""
    client = get_default_datature_client()
    artifacts = client.artifacts.list()

    artifact_choices = [
        Choice(name=f"{artifact.display_name} [{artifact.id[-6:]}]", value=artifact)
        for artifact in artifacts
    ]

    artifact = inquirer.fuzzy(
        message="Select a default model/artifact for this device:",
        choices=artifact_choices,
        max_height=10,
        border=True,
    ).execute()

    return artifact.id, artifact.display_name


def get_custom_config(
    deployment_id: str,
    secret_key: str,
    project_id: str,
    project_name: str,
    artifact_id: str,
    artifact_name: str,
) -> DeploymentConfig:
    """Get configuration for custom setup"""

    runtime_choices = [
        Choice(name="CPU", value="CPU"),
        Choice(name="GPU", value="GPU"),
    ]

    runtime = inquirer.select(
        message="Do you want to use CPU or GPU?",
        choices=runtime_choices,
        default="CPU",
        border=True,
    ).execute()

    gpu_count = 0
    if runtime == "GPU":
        gpu_count = int(
            inquirer.number(
                "How many GPUs do you want to use? (0 for all GPUs):",
                default=1,
            ).execute()
        )

    port = int(
        inquirer.number(
            "Which port do you want to use for the local deployment server?",
            default=9449,
        ).execute()
    )

    allow_origins_input = inquirer.text(
        message="Enter the allowed origins (comma-separated list):",
        default="https://nexus.datature.io",
    ).execute()

    # Clean up the origins list by stripping whitespace
    allow_origins = [
        origin.strip() for origin in allow_origins_input.split(",") if origin.strip()
    ]

    print()  # newline

    return DeploymentConfig(
        deployment_id=deployment_id,
        secret_key=secret_key,
        project_id=project_id,
        project_name=project_name,
        artifact_id=artifact_id,
        artifact_name=artifact_name,
        runtime=runtime,
        devices=gpu_count if runtime == "GPU" else "auto",
        port=port,
        allowed_origins=allow_origins,
    )


def prompt_config() -> DeploymentConfig:
    """Prompt the user for the local deploy configuration"""
    deployment_id = str(uuid4())[-6:]

    project = get_project_credentials()

    setup_choices = [
        Choice(name="Simple (Default Configuration)", value="simple"),
        Choice(name="Advanced (Custom Configuration)", value="custom"),
    ]

    setup_type = inquirer.select(
        message="Select the setup type:",
        choices=setup_choices,
        default="simple",
        border=True,
    ).execute()

    artifact_id, artifact_name = select_artifact()

    if setup_type == "simple":
        deployment_config = DeploymentConfig(
            deployment_id=deployment_id,
            secret_key=project.get("project_secret", ""),
            project_id=project.get("project_id", ""),
            project_name=project.get("project_name", ""),
            artifact_id=artifact_id,
            artifact_name=artifact_name,
        )

        print()  # newline
        print("Using default configuration...")
        print(f"Project: {deployment_config.project_name}")
        print(f"Runtime: {deployment_config.runtime}")
        print(f"Devices: {deployment_config.devices}")
        print(f"Port: {deployment_config.port}")
        print(f"Allowed Origins: {deployment_config.allowed_origins}")
        print()  # newline

        return deployment_config

    return get_custom_config(
        deployment_id=deployment_id,
        secret_key=project.get("project_secret", ""),
        project_id=project.get("project_id", ""),
        project_name=project.get("project_name", ""),
        artifact_id=artifact_id,
        artifact_name=artifact_name,
    )


def start_local_deployment():
    """Start the local deployment server."""
    deployment_configs = DeploymentConfigs.read_from_file()

    deployment_config = prompt_config()

    deployment_configs.configs.append(deployment_config)
    deployment_configs.write_to_file()

    print("Starting local deployment server...")

    try:
        start_server(
            deployment_id=deployment_config.deployment_id,
            secret_key=deployment_config.secret_key,
            project_id=deployment_config.project_id,
            project_name=deployment_config.project_name,
            artifact_id=deployment_config.artifact_id,
            artifact_name=deployment_config.artifact_name,
            runtime=deployment_config.runtime,
            devices=deployment_config.devices,
            port=deployment_config.port,
            allowed_origins=deployment_config.allowed_origins,
        )

    except KeyboardInterrupt:
        print("\nKeyboard interrupt detected. Shutting down...")

    except Exception as exc:
        print(f"An error occurred: {exc}")

    finally:
        deployment_configs = DeploymentConfigs.read_from_file()
        deployment_configs.configs.remove(deployment_config)
        deployment_configs.write_to_file()


def show_logs():
    """Show the logs of the local deployment server."""
    deployment_configs = DeploymentConfigs.read_from_file()

    if not deployment_configs.configs:
        print(
            "No local deployment server found. Please start a local deployment server "
            "first by running `datature local deploy start`."
        )
        sys.exit(1)

    deployment_choices = [
        Choice(
            name=f"{deployment_config.deployment_id} [Port: {deployment_config.port}]",
            value=deployment_config,
        )
        for deployment_config in deployment_configs.configs
    ]

    deployment = inquirer.select(
        message="Select the local deployment server to show logs for:",
        choices=deployment_choices,
        border=True,
    ).execute()

    try:
        url = f"http://localhost:{deployment.port}/all-logs"
        response = requests.get(url, timeout=REQUEST_TIME_OUT_SECONDS).json()

        for log in response["logs"]:
            timestamp = datetime.fromtimestamp(log["timestamp"] / 1000).strftime(
                "%Y-%m-%d %H:%M:%S.%f"
            )
            print(f"{timestamp}\t{log['message']}")

    except requests.exceptions.RequestException as exc:
        print(f"An error occurred: {exc}")
        sys.exit(1)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   label_map_pb2.py
@Author  :   Wei Loon Cheng
@Version :   1.0
@Contact :   hello@datature.io
@License :   Datature Outpost License 1.0
@Desc    :   Label map protocol buffer definition.

isort:skip_file
"""

# pylint: skip-file

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: label_map.proto
# Protobuf Python Version: 5.29.4
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 5, 29, 4, "", "label_map.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0flabel_map.proto\x12\x19\x64\x61tature.inference_server")\n\rLabelMapEntry\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t"B\n\x08LabelMap\x12\x36\n\x04item\x18\x01 \x03(\x0b\x32(.datature.inference_server.LabelMapEntry'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "label_map_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_LABELMAPENTRY"]._serialized_start = 46
    _globals["_LABELMAPENTRY"]._serialized_end = 87
    _globals["_LABELMAP"]._serialized_start = 89
    _globals["_LABELMAP"]._serialized_end = 155
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   device_config.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Outpost device configuration module.
"""

from pathlib import Path
from typing import Any, Dict, Union

import msgspec
import yaml
from msgspec import Struct, field

from datature.nexus.cli.outpost import consts as CONSTS


class DeviceConfig(Struct, kw_only=True):
    """Device configuration.

    :param device: Device configuration.
    :param authentication: Authentication configuration.
    :param management_api: Management API configuration.
    :param mqtt: MQTT configuration.
    :param statsd: StatsD configuration
    """

    device: Dict[str, Any] = field(default_factory=dict)
    authentication: Dict[str, Any] = field(default_factory=dict)
    management_api: Dict[str, Any] = field(default_factory=dict)
    mqtt: Dict[str, Any] = field(default_factory=dict)
    statsd: Dict[str, Any] = field(default_factory=dict)
    version: str = field(default="1.0.0")

    def write_to_file(self, config_path: Union[Path, str]) -> None:
        """Write service configuration to file."""
        try:
            config_dict = msgspec.to_builtins(self)

            with open(config_path, "w", encoding="utf-8") as config_file:
                yaml.dump(config_dict, config_file, indent=4, default_flow_style=False)

        except Exception as exc:
            raise ValueError(f"Error writing device configuration: {exc}") from exc


DEVICE_CONFIG = DeviceConfig(
    device={
        "name": "${{device_name}}",
        "id": "${{device_id}}",
        "workspace_id": "${{workspace_id}}",
        "config_root_dir": "DATATURE_OUTPOST_CONFIG_DIR",
    },
    authentication={
        "client_certificate": {
            "auto_renewal": {
                "days_before_expiry": 15,
                "issuer": "ManagementApi",
            },
            "certificate_file": "/home/datature/.datature/ssl/device.crt",
            "private_key_file": "/home/datature/.datature/ssl/key.pvt.pem",
        }
    },
    management_api={
        "authentication": {"kind": "ClientCertificate"},
        "endpoint": CONSTS.NEXUS_ENDPOINT,
    },
    mqtt={
        "authentication": {
            "issuing_ca": "/home/datature/.datature/ssl/issuing.ca",
            "kind": "ClientCertificate",
        },
        "endpoint": CONSTS.MQTT_ENDPOINT,
        "keep_alive_interval_seconds": 60,
        "protocol": "OutpostV1",
        "reconnect_delay_seconds": 5,
        "transport": "TLS",
    },
    statsd={"host": "localhost", "port": 8125},
    version="1.0.0",
)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   system_specs.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Outpost system specifications module.
"""

# pylint: disable=R0914,W0718,E1102

import queue
import subprocess
import sys
from typing import List

from datature.nexus.cli.outpost import consts as CONSTS
from datature.nexus.cli.outpost import messages as MESSAGES
from datature.nexus.cli.outpost.common import install_pip_package
from datature.nexus.cli.outpost.types import (
    Accelerator,
    AcceleratorMemory,
    AcceleratorPerformanceMetric,
    Camera,
    CameraCaptureProfile,
    CPUCore,
    CPUMemory,
    DeviceInfo,
    FrameRate,
    OutpostDeviceConfig,
)
from datature.nexus.cli.outpost.utils import start_spinner
from datature.nexus.error import Error

try:
    import psutil
except ImportError as exc:
    if not install_pip_package("psutil"):
        raise Error(
            MESSAGES.INSTALL_PIP_PACKAGE_ERROR_MESSAGE.format(package_name="psutil")
        ) from exc
    import psutil

try:
    import pyudev
except ImportError as exc:
    if not install_pip_package("pyudev"):
        raise Error(
            MESSAGES.INSTALL_PIP_PACKAGE_ERROR_MESSAGE.format(package_name="pyudev")
        ) from exc
    import pyudev

try:
    from linuxpy.video.device import BufferType, Device, V4L2Error
except ImportError as exc:
    if not install_pip_package("linuxpy"):
        raise Error(
            MESSAGES.INSTALL_PIP_PACKAGE_ERROR_MESSAGE.format(package_name="linuxpy")
        ) from exc
    from linuxpy.video.device import BufferType, Device, V4L2Error

try:
    import pynvml
except ImportError as exc:
    if not install_pip_package("nvidia-ml-py"):
        raise Error(
            MESSAGES.INSTALL_PIP_PACKAGE_ERROR_MESSAGE.format(
                package_name="nvidia-ml-py"
            )
        ) from exc
    import pynvml


def create_device_id():
    """
    Create unique device ID, which is a combination of 'outpostdevice_' and the last 22 characters
    of the machine ID.
    """
    with open("/etc/machine-id", "r", encoding="utf-8") as machine_id_file:
        device_id = f"outpostdevice_{str(machine_id_file.read().strip())[-22:]}"
    return device_id


def get_cameras_info():
    """Get camera information."""
    cameras = []
    try:
        context = pyudev.Context()
        for device in context.list_devices(subsystem="video4linux"):
            with Device(device.device_node) as video_input:
                try:
                    input_format = video_input.get_format(BufferType.VIDEO_CAPTURE)
                except OSError:
                    continue

                resolution = [input_format.width, input_format.height]
                frame_rate = FrameRate(
                    frames=int(video_input.get_fps(BufferType.VIDEO_CAPTURE)),
                    interval="Second",
                )

                capture_profile = CameraCaptureProfile(
                    resolution=resolution,
                    max_frame_rate=frame_rate,
                )

                camera = Camera(
                    id=str(device.device_node.strip("/dev/video")),
                    name=video_input.info.card,
                    capture_profiles=[capture_profile],
                )

                cameras.append(camera)

        if len(cameras) == 0:
            raise V4L2Error("No input devices found")

    except V4L2Error as exc:
        print(f"Error getting camera information: {exc}")
        print(
            "Warning: Please ensure that you have a camera connected to your device when starting the runtime."
        )

    except Exception as exc:
        print(f"Unknown error when getting camera information: {exc}")
        print(f"Please contact {CONSTS.SUPPORT_EMAIL} for assistance.")
        sys.exit(1)

    return cameras


def get_cpu_core_info():
    """Get CPU core information."""
    cpu_model_name = ""
    try:
        with open("/proc/cpuinfo", "r", encoding="utf-8") as f:
            for line in f:
                if "model name" in line:
                    cpu_model_name = line.split(":")[1].strip()
                    break
    except Exception:
        cpu_model_name = "Unknown CPU"

    core_count = psutil.cpu_count(logical=False)

    return CPUCore(name=cpu_model_name, count=core_count)


def get_cpu_memory_info():
    """Get CPU memory information, in bytes."""
    return CPUMemory(bytes=psutil.virtual_memory().total)


def get_disk_space_bytes():
    """Get disk space in bytes."""
    return int(psutil.disk_usage("/").total)


def get_accelerator_info(device_model_name: str) -> List[Accelerator]:
    """Get accelerator information for CUDA devices."""
    try:
        pynvml.nvmlInit()

        driver_version = pynvml.nvmlSystemGetDriverVersion()
        if isinstance(driver_version, bytes):
            driver_version = driver_version.decode("utf-8")

        cuda_version = pynvml.nvmlSystemGetCudaDriverVersion()
        device_count = pynvml.nvmlDeviceGetCount()

        accelerators = []

        for i in range(device_count):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            device_name = str(pynvml.nvmlDeviceGetName(handle))

            try:
                device_vram_bytes = int(pynvml.nvmlDeviceGetMemoryInfo(handle).total)

            except pynvml.NVMLError as exc:
                if "jetson" in device_model_name.lower():
                    device_vram_bytes = get_cpu_memory_info().bytes
                else:
                    raise exc from exc

            device_compute_capability = pynvml.nvmlDeviceGetCudaComputeCapability(
                handle
            )

            device_uuid = pynvml.nvmlDeviceGetUUID(handle)
            if isinstance(device_uuid, bytes):
                device_uuid = device_uuid.decode("utf-8")

            num_cuda_cores = pynvml.nvmlDeviceGetNumGpuCores(handle)

            accelerator_memory = AcceleratorMemory(
                bytes=device_vram_bytes,
                shared=bool("jetson" in device_model_name.lower()),
            )

            accelerator_performance_metrics = [
                AcceleratorPerformanceMetric(
                    name="CUDA Cores",
                    value=num_cuda_cores,
                ),
                AcceleratorPerformanceMetric(
                    name="Compute Capability",
                    value=".".join(map(str, device_compute_capability)),
                ),
                AcceleratorPerformanceMetric(
                    name="CUDA Version",
                    value=cuda_version,
                ),
                AcceleratorPerformanceMetric(
                    name="Driver Version",
                    value=driver_version,
                ),
            ]

            accelerator = Accelerator(
                manufacturer="NVIDIA",
                name=device_name,
                id=device_uuid.replace("GPU-", ""),
                memory=accelerator_memory,
                performance_metrics=accelerator_performance_metrics,
            )

            accelerators.append(accelerator)

        pynvml.nvmlShutdown()

        return accelerators

    except subprocess.CalledProcessError as exc:
        raise Error(f"Error running nvidia-settings: {exc.stderr}") from exc

    except pynvml.NVMLError:
        print("\nNo CUDA support found, skipping accelerator check.")

    except Error as exc:
        raise Error(f"Error getting CUDA accelerator information: {exc}") from exc

    return []


def get_device_info():
    """Get device information."""
    try:
        with open(
            "/sys/devices/virtual/dmi/id/product_name", "r", encoding="utf-8"
        ) as f:
            device_model_name = f.read().strip()
    except Exception as exc:
        print(f"Error getting device model name: {exc}, using 'Unknown Device'")
        device_model_name = "Unknown Device"

    # Initialize variables with default values
    cpu_core = None
    cpu_memory = None
    disk_space_bytes = 0
    cameras = []
    accelerators = []

    try:
        cpu_core = get_cpu_core_info()
        cpu_memory = get_cpu_memory_info()
        disk_space_bytes = get_disk_space_bytes()
        cameras = get_cameras_info()
        accelerators = get_accelerator_info(device_model_name)

    except Exception as exc:
        print(f"Error getting device information: {exc}")

    # Ensure we have valid values before returning
    if cpu_core is None:
        cpu_core = CPUCore(name="Unknown CPU", count=1)
    if cpu_memory is None:
        cpu_memory = CPUMemory(bytes=0)

    return (
        device_model_name,
        cpu_core,
        cpu_memory,
        disk_space_bytes,
        cameras,
        accelerators,
    )


def get_cameras_in_use():
    """Get cameras in use."""
    in_use_cameras = []

    try:
        context = pyudev.Context()
        for device in context.list_devices(subsystem="video4linux"):
            with Device(device.device_node) as video_input:
                # Try to start streaming (will fail if device is busy)
                try:
                    for _ in video_input:
                        break
                except IOError as io_exc:
                    if "Device or resource busy" in str(io_exc):
                        device_id = str(device.device_node.replace("/dev/video", ""))
                        in_use_cameras.append(device_id)
                except V4L2Error:
                    continue

    except V4L2Error as exc:
        print(f"Error: {exc}")
        sys.exit(1)

    return in_use_cameras


def get_current_device_info():
    """Get current device information."""
    cpu_cores_used = float(
        psutil.cpu_count(logical=True) * psutil.cpu_percent(interval=1) / 100
    )
    cpu_memory_used_bytes = int(psutil.virtual_memory().used)
    disk_space_used_bytes = int(psutil.disk_usage("/").used)
    cameras_in_use = get_cameras_in_use()

    return {
        "cpu_cores_used": cpu_cores_used,
        "cpu_memory_used_bytes": cpu_memory_used_bytes,
        "disk_space_used_bytes": disk_space_used_bytes,
        "cameras_in_use": cameras_in_use,
    }


@start_spinner
def check_system_specs(
    outpost_device_config: OutpostDeviceConfig, message_queue: queue.Queue
) -> None:
    """Check system specifications.

    :param outpost_device_config (OutpostDeviceConfig): Outpost device configuration.
    :param message_queue (queue.Queue): Message queue.
    """
    message_queue.put("Checking system specifications...")
    try:
        outpost_device_config.id = create_device_id()
        (
            device_model_name,
            cpu_core,
            cpu_memory,
            disk_space_bytes,
            cameras,
            accelerators,
        ) = get_device_info()

    except Exception as exc:
        print(f"Error getting device information: {exc}")
        print(
            "Please ensure that you are running this command on a supported Linux machine."
        )
        print(
            f"The list of supported machines can be found at: {CONSTS.DOCS_ENV_SETUP_URL}"
        )
        print(
            "If you wish to run this command on an unsupported machine,"
            f" please contact us at {CONSTS.SUPPORT_EMAIL}."
        )
        sys.exit(1)

    if cpu_memory.bytes < CONSTS.MINIMUM_RAM_BYTES:
        raise ValueError(
            f"Insufficient RAM. You have {cpu_memory.bytes} bytes,"
            f" but you need at least {CONSTS.MINIMUM_RAM_BYTES} bytes"
        )

    if disk_space_bytes < CONSTS.MINIMUM_STORAGE_BYTES:
        raise ValueError(
            f"Insufficient storage space. You have {disk_space_bytes}"
            f" bytes, but you need at least {CONSTS.MINIMUM_STORAGE_BYTES} bytes"
        )

    device_info = DeviceInfo(
        name=device_model_name,
        components=[cpu_core, cpu_memory] + cameras + accelerators,
    )

    outpost_device_config.device_info = device_info
    outpost_device_config.write_to_file()

    return "Success: System requirements verified."



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   types.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Outpost types.
"""

from __future__ import annotations

import json
import sys
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import msgspec
from msgspec import MsgspecError, Struct, field

from datature.nexus.cli.outpost.consts import OUTPOST_CONFIG_FILE_PATH


class DateTimeEncoder(json.JSONEncoder):
    """JSON encoder for datetime objects."""

    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()
        return json.JSONEncoder.default(self, o)


class ConfigVariable(Struct, rename="camel", kw_only=True):
    """Config variable object."""

    kind: str = ""
    role: str = ""
    default: str = ""


class OutpostDeviceConfigurationSpec(Struct, rename="camel", kw_only=True):
    """Outpost configuration spec struct.

    :param kind: Configuration kind.
    :param config_version: Configuration version.
    :param config_data: Configuration data.
    :param config_variables: Configuration variables.
    """

    kind: str = ""
    config_version: str = ""
    config_data: Dict[str, str] = field(default_factory=dict)
    config_variables: Dict[str, ConfigVariable] = field(default_factory=dict)


class OutpostDeviceConfiguration(Struct, rename="camel", kw_only=True):
    """Outpost configuration struct.

    :param id: Configuration ID.
    :param object: Configuration object type.
    :param workspace_id: Workspace ID.
    :param name: Configuration name.
    :param spec: Configuration spec.
    :param create_date: Configuration creation date.
    :param update_date: Configuration update date.
    :param metadata_generation: Configuration metadata generation.
    """

    id: str = ""
    object: str = ""
    workspace_id: str = ""
    name: str = ""
    spec: OutpostDeviceConfigurationSpec = field(
        default_factory=OutpostDeviceConfigurationSpec
    )
    create_date: int = 0
    update_date: int = 0
    metadata_generation: int = 0


class FrameRate(
    Struct,
    omit_defaults=True,
    rename="camel",
    kw_only=True,
):
    """Frame rate struct.

    :param frames: Number of frames to take in the interval.
    :param interval: Time interval for measuring frame rate in seconds.
    """

    frames: int
    interval: str


class CameraCaptureProfile(
    Struct,
    omit_defaults=True,
    rename="camel",
    kw_only=True,
):
    """Camera capture profile struct.

    :param resolution: Pixel resolution of the camera in [width, height], e.g., [1920, 1080].
    :param max_frame_rate: Frame rate of the camera.
    """

    resolution: List[int]
    max_frame_rate: FrameRate


class CPUCore(
    Struct,
    omit_defaults=True,
    rename="camel",
    tag="CpuCore",
    tag_field="kind",
    kw_only=True,
):
    """CPU core struct.

    :param name: Core name.
    :param count: Number of cores.
    """

    name: str
    count: int


class CPUMemory(
    Struct,
    omit_defaults=True,
    rename="camel",
    tag="CpuMemory",
    tag_field="kind",
    kw_only=True,
):
    """CPU memory struct.

    :param bytes: Total memory in bytes.
    """

    bytes: int


class AcceleratorMemory(Struct, omit_defaults=True, rename="camel", kw_only=True):
    """Accelerator memory struct.

    :param bytes: Total memory in bytes.
    :param shared: Whether the memory is shared.
    """

    bytes: int
    shared: bool


class AcceleratorPerformanceMetric(
    Struct, omit_defaults=True, rename="camel", kw_only=True
):
    """Accelerator performance metric struct.

    :param name: Name of the performance metric.
    :param value: Value of the performance metric.
    """

    name: str
    value: int | float | str


class Accelerator(
    Struct,
    omit_defaults=True,
    rename="camel",
    tag="Accelerator",
    tag_field="kind",
    kw_only=True,
):
    """Accelerator struct.

    :param manufacturer: Manufacturer of the accelerator.
    :param name: Name of the accelerator.
    :param id: Accelerator ID.
    :param memory: Accelerator memory.
    :param performance_metrics: List of performance metrics for the accelerator.
    """

    manufacturer: str
    name: str
    id: str
    memory: AcceleratorMemory
    performance_metrics: list[AcceleratorPerformanceMetric]


class Camera(
    Struct,
    omit_defaults=True,
    rename="camel",
    tag="Camera",
    tag_field="kind",
    kw_only=True,
):
    """Camera struct.

    :param id: Camera ID used for capturing images (this is typically 0, 1, 2, etc.). Default is -1.
    :param name: Camera name or brand.
    :param capture_profiles: List of camera capture profiles.
    """

    id: str
    name: str
    capture_profiles: List[CameraCaptureProfile]


DeviceComponent = Union[CPUCore, CPUMemory, Camera, Accelerator]


class DeviceInfo(Struct, omit_defaults=True, rename="camel", kw_only=True):
    """Device information struct.

    :param name: Device name.
    :param components: List of device components.
    """

    name: str = ""
    components: List[DeviceComponent] = field(default_factory=list)


class RegistrationStatus(Enum):
    """Registration status enum."""

    ARCHIVED = "Archived"
    FROZEN = "Frozen"
    PAUSED = "Paused"
    REGISTERED = "Registered"
    TESTING = "Testing"


class DeviceStatusMessage(Enum):
    """Device status message enum."""

    ARCHIVED = "Archived"
    ARCHIVED_ROLLING_OUT = "ArchivedRollingOut"
    FROZEN = "Frozen"
    FROZEN_ROLLING_OUT = "FrozenRollingOut"
    OFFLINE = "Offline"
    PAUSED = "Paused"
    PAUSED_ROLLING_OUT = "PausedRollingOut"
    READY = "Ready"
    READY_FAILED_ROLL_OUT = "ReadyFailedRollOut"
    READY_ROLLING_OUT = "ReadyRollingOut"
    REGISTERING = "Registering"


class DeviceStatusProperties(Struct, omit_defaults=True, rename="camel", kw_only=True):
    """Device status properties struct.

    :param message: Status message.
    :param last_connected_at: Last connected timestamp.
    :param sample_prediction: Sample prediction data.
    """

    message: Optional[str] = None
    last_connected_at: Optional[int] = None
    sample_prediction: Optional[Dict[str, Any]] = None


class UpdaterStatus(
    Struct,
    omit_defaults=True,
    rename="camel",
    kw_only=True,
):
    """Updater status struct.

    :param status: Updater status.
    :param pid: Process ID of the updater.
    :param status_output: Output of the updater status.
    """

    status: Optional[str] = None
    pid: Optional[int] = None
    status_output: Optional[str] = None


class RuntimeStatus(
    Struct,
    omit_defaults=True,
    rename="camel",
    kw_only=True,
):
    """Runtime status struct.

    :param status: Runtime status.
    :param pid: Process ID of the runtime.
    :param status_output: Output of the runtime status.
    """

    status: Optional[str] = None
    pid: Optional[int] = None
    status_output: Optional[str] = None


class DeviceStatus(
    Struct,
    omit_defaults=True,
    rename="camel",
    kw_only=True,
):
    """Device status struct.

    :param status: Device status.
    :param updater_status: Updater status.
    :param runtime_status: Runtime status.
    :param last_updated: Last updated timestamp.
    :param valid_certificate: Whether the certificate is valid.
    :param certificate_expires_at: Certificate expiration timestamp.
    """

    status: DeviceStatusProperties = field(default_factory=DeviceStatusProperties)
    updater_status: UpdaterStatus = field(default_factory=UpdaterStatus)
    runtime_status: RuntimeStatus = field(default_factory=RuntimeStatus)
    last_updated: Optional[datetime] = None
    valid_certificate: bool = False
    certificate_expires_at: Optional[datetime] = None


class DeviceCredentials(
    Struct,
    omit_defaults=True,
    rename="camel",
    kw_only=True,
):
    """Device credentials struct.

    :param secret_key: Workspace secret key.
    :param workspace_id: Workspace ID.
    :param default_project_id: Default project ID.
    """

    secret_key: str = ""
    workspace_id: str = ""
    default_project_id: str = ""


class RuntimeVersion(Struct, rename="camel", kw_only=True):
    """Runtime version struct.

    :param major: Major version.
    :param minor: Minor version.
    :param patch: Patch version.
    """

    major: int = 0
    minor: int = 0
    patch: int = 0

    def __str__(self):
        return f"{self.major}.{self.minor}.{self.patch}"


class RuntimeLink(Struct, rename="camel", kw_only=True):
    """Runtime link struct.

    :param url: Signed URL of the runtime zip file.
    :param expires_at: Expiration timestamp of the link.
    """

    url: str = ""
    expires_at: int = 0


class OutpostDeviceRuntime(Struct, rename="camel", kw_only=True):
    """Outpost runtime struct.

    :param id: Runtime ID.
    :param object: Runtime object type.
    :param build: Runtime build.
    :param version: Runtime version.
    :param link: Runtime link.
    :param description: Runtime description.
    :param is_visible: Whether the runtime is visible.
    :param yanked: Whether the runtime is yanked.
    :param create_date: Runtime creation date.
    :param update_date: Runtime update date.
    :param metadata_generation: Runtime metadata generation.
    """

    id: str = ""
    object: str = ""
    build: str = ""
    version: RuntimeVersion = field(default_factory=RuntimeVersion)
    link: RuntimeLink = field(default_factory=RuntimeLink)
    description: str = ""
    is_visible: bool = True
    yanked: bool = False
    create_date: int = 0
    update_date: int = 0
    metadata_generation: int = 0


class MetadataSpec(Struct, rename="camel", kw_only=True):
    """Metadata spec struct.

    :param device_uid: Device UID.
    :param tags: Device tags.
    :param metadata_generation: Metadata generation.
    :param configuration: Device configuration.
    :param runtime: Device runtime.
    """

    device_uid: str = ""
    tags: Dict[str, str] = field(default_factory=dict)
    metadata_generation: int = 0
    revision_id: str = ""
    configuration: OutpostDeviceConfiguration = field(
        default_factory=OutpostDeviceConfiguration
    )
    runtime: OutpostDeviceRuntime = field(default_factory=OutpostDeviceRuntime)
    runtime_variables: Dict[str, Any] = field(default_factory=dict)


class OutpostDeviceConfig(
    Struct,
    omit_defaults=True,
    rename="camel",
    kw_only=True,
):
    """Outpost device configuration struct.

    :param id: Device ID.
    :param uid: Device UID.
    :param name: Device name.
    :param tags: Device tags.
    :param default_artifact_id: Default artifact ID.
    :param metadata_generation: Metadata generation.
    :param configuration: Device configuration.
    :param runtime: Device runtime.
    :param runtime_dir: Runtime directory.
    :param device_info: Device information.
    :param device_status: Device status.
    :param device_credentials: Device credentials.
    """

    id: str = ""
    uid: str = ""
    name: str = ""
    tags: Dict[str, str] = field(default_factory=dict)
    default_artifact_id: str = ""
    metadata_generation: int = 0
    configuration: OutpostDeviceConfiguration = field(
        default_factory=OutpostDeviceConfiguration
    )
    runtime: OutpostDeviceRuntime = field(default_factory=OutpostDeviceRuntime)
    runtime_dir: str = ""
    device_info: DeviceInfo = field(default_factory=DeviceInfo)
    device_status: DeviceStatus = field(default_factory=DeviceStatus)
    device_credentials: DeviceCredentials = field(default_factory=DeviceCredentials)

    def __post_init__(self):
        if not OUTPOST_CONFIG_FILE_PATH.exists():
            with open(OUTPOST_CONFIG_FILE_PATH, "w", encoding="utf-8"):
                pass

    def write_to_file(self) -> None:
        """Write Outpost configuration to file."""
        try:
            outpost_device_config_dict = msgspec.to_builtins(self)

            with open(OUTPOST_CONFIG_FILE_PATH, "w", encoding="utf-8") as config_file:
                json.dump(
                    outpost_device_config_dict,
                    config_file,
                    cls=DateTimeEncoder,
                    indent=4,
                )

        except (IOError, MsgspecError) as exc:
            print(f"Error writing Outpost config to {OUTPOST_CONFIG_FILE_PATH}: {exc}")
            sys.exit(1)

    @classmethod
    def read_from_file(cls) -> OutpostDeviceConfig:
        """Read Outpost configuration from file."""
        if not OUTPOST_CONFIG_FILE_PATH.exists():
            print(
                "Device not registered. "
                "Please run 'datature outpost install' to register this device."
            )
            sys.exit(1)

        try:
            if OUTPOST_CONFIG_FILE_PATH.stat().st_size == 0:
                return cls()

            with open(OUTPOST_CONFIG_FILE_PATH, "r", encoding="utf-8") as config_file:
                json_data = json.loads(config_file.read())
                return msgspec.convert(json_data, type=OutpostDeviceConfig)

        except (IOError, MsgspecError) as exc:
            print(
                f"Error reading Outpost config from {OUTPOST_CONFIG_FILE_PATH}: {exc}"
            )
            sys.exit(1)

    def remove_from_file(self) -> None:
        """Remove Outpost configuration from file."""

        with open(OUTPOST_CONFIG_FILE_PATH, "w", encoding="utf-8") as config_file:
            config_file.write("")



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   helper.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Common helper functions for Outpost.
"""

# pylint: disable=W0719

import queue
import re
import sys
import threading
from functools import wraps
from typing import Dict

from colorama import Fore
from halo import Halo

from datature.nexus.error import Error

VAR_RE = re.compile(
    r"\${{\s*([a-zA-Z_][a-zA-Z0-9_]{0,63})\s*}}", flags=re.DOTALL | re.MULTILINE
)


def start_spinner(func: callable = None, raise_on_error: bool = False) -> callable:
    """Start spinner for the function."""

    def decorator(func: callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            wait_spinner = Halo(spinner="dots")
            message_queue = queue.Queue()
            stop_thread = threading.Event()

            def update_spinner():
                while not stop_thread.is_set():
                    try:
                        new_message = message_queue.get(timeout=0.1)
                        if not kwargs.get("quiet", False) and new_message is not None:
                            wait_spinner.text = new_message
                    except queue.Empty:
                        continue

            thread = threading.Thread(target=update_spinner)
            wait_spinner.start()
            thread.start()

            try:
                result = func(*args, **kwargs, message_queue=message_queue)
                if not kwargs.get("quiet", False):
                    wait_spinner.succeed(result)
                return result

            except Error as exc:
                error_symbol = f"{Fore.CYAN}●{Fore.RESET}"
                wait_spinner.stop_and_persist(error_symbol, exc.message)

                if raise_on_error:
                    raise exc
                sys.exit(1)

            except Exception as exc:  # pylint: disable=W0718
                error_symbol = f"{Fore.CYAN}●{Fore.RESET}"
                wait_spinner.stop_and_persist(error_symbol, str(exc))

                if raise_on_error:
                    raise exc
                sys.exit(1)

            finally:
                stop_thread.set()
                thread.join()
                wait_spinner.stop()

        return wrapper

    if func is None:
        return decorator
    return decorator(func)


def interpolate_config_variables(config_str: str, variables: Dict[str, str]) -> str:
    """Replace all occurrences of ${{ variable_name }} in config_str with
    corresponding values from the variables dictionary.
    """

    def replace_var(match):
        var_name = match.group(1)
        if var_name not in variables:
            raise KeyError(f"Undefined variable referenced in template: '{var_name}'")

        return str(variables[var_name])

    # Use re.sub with a replacement function to replace all matches
    return VAR_RE.sub(replace_var, config_str)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   skeleton.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Keypoint skeleton types.
"""
# pylint: disable=C0103

from __future__ import annotations

from dataclasses import dataclass, field, replace
from typing import TypedDict
from uuid import uuid4


class SkeletonParsingError(ValueError):
    """An error occurred while parsing a skeleton."""

    def __init__(self, message: str):
        super().__init__(f"Error parsing skeleton: {message}")


class KeypointJson(TypedDict):
    """Keypoint json type."""

    id: str
    name: str
    category: list[str]


class ConnectionJson(TypedDict):
    """Connection json type."""

    pair: list[str]  # keypoint names


class SkeletonJson(TypedDict):
    """Skeleton json type."""

    skeletonId: str
    name: str
    keypoints: list[KeypointJson]
    connections: list[ConnectionJson]


@dataclass
class Keypoint:
    """Keypoint object."""

    name: str
    id_: str = field(default_factory=lambda: str(uuid4()))
    category: list[str] = field(default_factory=list)

    def __post_init__(self):
        if not isinstance(self.name, str):
            raise SkeletonParsingError(
                f"The keypoint name "
                f"{self.name} is not a string. "
                "Please ensure all keypoint names are strings."
            )

        if not isinstance(self.category, list):
            raise SkeletonParsingError(
                f"The keypoint category "
                f"{self.category} is not a list. "
                "Please ensure all keypoint categories are lists."
            )

        for category in self.category:
            if not isinstance(category, str):
                raise SkeletonParsingError(
                    f"The keypoint category "
                    f"{self.category} is not a list of strings. "
                    "Please ensure all keypoint categories are lists of "
                    "strings."
                )

    def to_json(self):
        """Convert keypoint object to json."""
        return {
            "name": self.name,
            "category": self.category,
        }

    @classmethod
    def from_json(cls, input_json: KeypointJson):
        """Create a keypoint object from a keypoint json."""
        if "name" not in input_json:
            raise SkeletonParsingError(
                "The keypoint "
                f"{input_json} does not have a 'name' field. "
                "Please ensure all keypoints have a "
                f"'name' field."
            )

        id_ = input_json.get("id", str(uuid4()))
        name = input_json["name"]
        category = input_json.get("category", [])

        return cls(
            id_=id_,
            name=name,
            category=category,
        )


@dataclass
class Connection:
    """Connection object."""

    pair: list  # keypoint names / indices

    def __post_init__(self):
        if not isinstance(self.pair, list):
            raise SkeletonParsingError(
                f"A given keypoint connection pair "
                f"{self.pair} is not a list. "
                "Please ensure all keypoint connection pairs are lists."
            )

        if len(self.pair) != 2:
            raise SkeletonParsingError(
                f"A given keypoint connection pair "
                f"{self.pair} is not a list of length 2. "
                "Please ensure all keypoint connection pairs are lists of "
                "length 2."
            )

        for name in self.pair:
            if not isinstance(name, (str, int)):
                raise SkeletonParsingError(
                    f"A given keypoint connection pair "
                    f"{self.pair} is not a list of strings or ints."
                    "Please ensure all keypoint connection pairs are "
                    "lists of strings."
                )

    def to_json(self):
        """Convert connection object to json."""
        return {
            "pair": self.pair,
        }

    @classmethod
    def from_json(cls, input_json):
        """Create a connection object from a connection json."""
        if not isinstance(input_json, dict):
            raise TypeError(
                f"The keypoint connection "
                f"{input_json} is not a dictionary. "
                "Please ensure all keypoint connections are dictionaries."
            )

        if "pair" not in input_json:
            raise SkeletonParsingError(
                "The keypoint connection "
                f"{input_json} does not have a 'pair' field. "
                "Please ensure all keypoint connections have a 'pair' field."
            )

        pair = input_json["pair"]
        return cls(pair=pair)


@dataclass
class Skeleton:
    """Skeleton object."""

    name: str
    keypoints: list[Keypoint]
    connections: list[Connection]

    @classmethod
    def from_json(cls, json: SkeletonJson):
        """Create a skeleton object from a skeleton json.

        Performs type checking to ensure sanity of skeleton.
        If connection key is "id", the connections are converted to named
        connections.
        """
        if not isinstance(json, dict):
            raise TypeError(
                "The skeleton json "
                "provided is not a dictionary. "
                "Please ensure all skeleton jsons are dictionaries."
            )

        for skeleton_field in ["name", "connections"]:
            if skeleton_field not in json:
                raise SkeletonParsingError(
                    "The skeleton "
                    f"provided does not have a '{skeleton_field}' field. "
                    "Please ensure all skeletons have a "
                    f"{skeleton_field} field."
                )

        name = json["name"]
        if not isinstance(name, str):
            raise SkeletonParsingError(
                "The skeleton name "
                f"{name} is not a string. "
                "Please ensure all skeleton names are strings."
            )

        return cls(
            name=json["name"],
            keypoints=[Keypoint.from_json(kp) for kp in json["keypoints"]],
            connections=[Connection.from_json(con) for con in json["connections"]],
        )

    def index_connections(self, offset=0) -> list[Connection]:
        """Convert named connections to index connections.

        Args:
            offset: The offset to add to the indices.
            NOTE: this is used for coco reupload because it starts at 1.
        """
        keypoint_names = [kp.name for kp in self.keypoints]

        return [
            replace(con, pair=[keypoint_names.index(i) + offset for i in con.pair])
            for con in self.connections
        ]

    def compare(self, other: Skeleton) -> str:
        """Compare two skeletons and return differences."""
        differences = ""

        if not isinstance(other, Skeleton):
            differences += "Skeletons are not the same type.\n"

        if self.name != other.name:
            differences += (
                "Skeletons have different names " f"{self.name} and {other.name}.\n"
            )

        self_kp_names = [kp.name for kp in self.keypoints]
        other_kp_names = [kp.name for kp in other.keypoints]

        for position, (self_kp_name, other_kp_name) in enumerate(
            zip(self_kp_names, other_kp_names)
        ):
            if self_kp_name != other_kp_name:
                differences += (
                    f"Keypoint at position {position} have different names "
                    f"{self_kp_name} and "
                    f"{other_kp_name}.\n"
                )

        self_connections = [
            frozenset(connection.pair) for connection in self.connections
        ]

        other_connections = [
            frozenset(connection.pair) for connection in other.connections
        ]

        self_connections.sort(key=hash)
        other_connections.sort(key=hash)

        for self_connection, other_connection in zip(
            self_connections, other_connections
        ):
            if self_connection != other_connection:
                differences += (
                    f"Connection {self_connection} and {other_connection} "
                    "are not the same.\n"
                )

        if len(differences) > 0:
            differences = differences[:-1]  # Remove last newline
        return differences

    def get_ids(self, other: Skeleton) -> list[int]:
        """Get the indices of the keypoints in the other skeleton."""
        self_kp_names = [kp.name for kp in self.keypoints]
        indices = [self_kp_names.index(kp.name) for kp in other.keypoints]

        return indices

    def to_json(self):
        """Convert skeleton object to json."""
        json_keypoints = [keypoint.to_json() for keypoint in self.keypoints]
        json_connections = [connection.to_json() for connection in self.connections]

        return {
            "name": self.name,
            "keypoints": json_keypoints,
            "connections": json_connections,
        }



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   endpoint_response.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Endpoint response type.
"""

from typing import List

from msgspec import Struct


class HealthResponse(Struct, kw_only=True, omit_defaults=True, rename="camel"):
    """Health response object.

    Attributes:
        project: The current project ID.
        project_name: The current project name.
        timestamp: The timestamp of the response in milliseconds.
        status: The status of the response.
    """

    project: str
    project_name: str
    timestamp: int
    status: str


class LogEntry(Struct, kw_only=True, omit_defaults=True, rename="camel"):
    """Log entry object.

    Attributes:
        timestamp: The timestamp of the log entry in milliseconds.
        message: The message of the log entry.
    """

    timestamp: int
    message: str


class LogResponse(Struct, kw_only=True, omit_defaults=True, rename="camel"):
    """Log response object.

    Attributes:
        logs: The logs of the response.
    """

    logs: List[LogEntry]



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   prediction.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Prediction API
"""

from typing import Union

from datature.nexus import models
from datature.nexus.api.types import Pagination, PredictionFilter
from datature.nexus.client_context import RestContext


class Prediction(RestContext):
    """Datature Prediction API Resource."""

    def list(
        self,
        pagination: Union[Pagination, dict, None] = None,
        filters: Union[PredictionFilter, dict, None] = None,
    ) -> models.PaginationResponse[models.Prediction]:
        """Lists predictions in the project.

        :param pagination: Pagination object.
        :param filters: PredictionFilter object.

        :return: PaginationResponse containing a list of predictions with the following structure:

                .. code-block:: python

                    PaginationResponse(
                        next_page='NjVhYTE0ZDY3YzQ5MThlMTJjYTUy',
                        prev_page=None,
                        data=[
                            Prediction(
                                predictions=[
                                    PredictionObject(
                                        id='prediction_1d5f2062-0638-41d4-8317-b46f226b7ee9:241',
                                        object='prediction',
                                        prediction_slice=PredictionSlice(
                                            project='proj_ca5fe71e7592bbcf7705ea36e4f29ed4',
                                            asset_id='asset_facf5eee-89a8-41f4-a958-95d54708e590',
                                            filename='29.jpg',
                                            slice=[],
                                            snapshot_id='snapshot_1d5f2062-0638-41d4-8317-b46f226b7ee9',
                                            run_id='run_4eecc30d-ac18-4821-96c5-239e97e96d6c',
                                            artifact_id='artifact_67a9a6fe9841727bfcc9c847',
                                            id='predictionslice_adc613a5-aa32-4c0f-86a1-5d5637c3bd26',
                                            prediction_metadata=PredictionSliceMetadata(
                                                average_confidence=0.79,
                                                average_confidence_for_tag=[
                                                'RBC',
                                                'WBC',
                                                'Platelets'
                                                ],
                                                prediction_count=26,
                                                prediction_count_for_tag=[
                                                'Background',
                                                'RBC',
                                                'Platelets',
                                                'WBC'
                                                ]
                                            )
                                        ),
                                        metadata=PredictionMetadata(
                                            average_confidence=0.25
                                        ),
                                        tag=Tag(
                                            index=1,
                                            name='RBC',
                                            color='#38b6ff',
                                            description=None
                                        ),
                                        bound=[
                                            'boundType',
                                            'rectangle'
                                        ],
                                    )
                                ],
                            )
                        ]

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.predictions.list(
                            filters={
                                "prediction_filter": '''
                                {
                                    "and": [
                                        {
                                        "=": [
                                            {
                                            "var": "run"
                                            },
                                            "run_4eecc30d-ac18-4821-96c5-239e97e96d6c"
                                        ]
                                        }
                                    ]
                                }
                                '''
                            }
                        )
        """
        assert isinstance(pagination, (Pagination, dict, type(None)))
        assert isinstance(filters, (PredictionFilter, dict, type(None)))

        if isinstance(pagination, dict):
            pagination = Pagination(**pagination)
        if pagination is None:
            pagination = Pagination()

        if isinstance(filters, dict):
            filters = PredictionFilter(**filters)
        if filters is None:
            filters = PredictionFilter()

        return self.requester.GET(
            f"/projects/{self.project_id}/predictions",
            query={**pagination.to_json(), **filters.to_json()},
            response_type=models.PaginationResponse[models.Prediction],
        )
